#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import config
from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import network_off
from earthkit.data.testing import write_to_file


@pytest.mark.skipif(  # TODO: fix
    sys.platform == "win32",
    reason="file:// not working on Windows yet",
)
def test_url_file_source():
    filename = os.path.abspath(earthkit_examples_file("test.nc"))
    s = from_source("url", f"file://{filename}")
    assert len(s) == 2


def test_url_source_1():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
    )


@pytest.mark.cache
def test_url_source_check_out_of_date():
    def load():
        from_source(
            "url",
            "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
        )

    with temp_directory() as tmpdir:
        with config.temporary():
            config.set("user-cache-directory", tmpdir)
            load()

            config.set("check-out-of-date-urls", False)
            with network_off():
                load()


def test_url_source_2():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
    )


def test_url_source_3():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.nc",
    )


def test_url_source_tar():
    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test_gribs.tar",
    )
    assert len(ds) == 6


def test_parts_url():
    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
    )

    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
        parts=((0, 4),),
    )

    assert os.path.getsize(ds.path) == 4

    with open(ds.path, "rb") as f:
        assert f.read() == b"BUFR"

    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
        parts=((0, 10), (50, 10), (60, 10)),
    )

    print(ds.path)

    assert os.path.getsize(ds.path) == 30

    with open(ds.path, "rb") as f:
        assert f.read()[:4] == b"BUFR"


def test_parts_as_arg_url_1():
    ds = from_source(
        "url",
        [
            "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
            [(0, 4)],
        ],
    )

    assert os.path.getsize(ds.path) == 4

    with open(ds.path, "rb") as f:
        assert f.read() == b"BUFR"


def test_parts_as_arg_url_2():
    ds = from_source(
        "url",
        [
            "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
            None,
        ],
    )

    assert os.path.getsize(ds.path) > 4

    with open(ds.path, "rb") as f:
        assert f.read(4) == b"BUFR"


def test_multi_url_parts_as_arg_invalid_1():
    with pytest.raises(ValueError):
        from_source(
            "url",
            [
                "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
                [(0, 4)],
            ],
            parts=[(0, 5)],
        )


def test_multi_url_parts_invalid():
    parts1 = [(240, 150)]
    parts2 = [(0, 526)]
    with pytest.raises(ValueError):
        from_source(
            "url",
            [
                [earthkit_remote_test_data_file("examples/test6.grib"), parts1],
                [earthkit_remote_test_data_file("examples/test.grib"), parts2],
            ],
            parts=[(0, 240)],
        )


@pytest.mark.skipif(  # TODO: fix
    sys.platform == "win32",
    reason="file:// not working on Windows yet",
)
def test_url_part_file_source():
    filename = os.path.abspath(earthkit_examples_file("test.grib"))
    ds = from_source(
        "url",
        f"file://{filename}",
        parts=[
            (0, 4),
            (522, 4),
            (526, 4),
            (1048, 4),
        ],
    )

    assert os.path.getsize(ds.path) == 16

    with open(ds.path, "rb") as f:
        assert f.read() == b"GRIB7777GRIB7777"


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_url_netcdf_source_save(write_method):
    ds = from_source(
        "url",
        earthkit_remote_test_data_file("examples/test.nc"),
    )

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        assert os.path.exists(tmp)
        ds_saved = from_source("file", tmp)
        assert len(ds_saved) == 2


if __name__ == "__main__":
    import logging

    from earthkit.data.testing import main

    logging.basicConfig(level=logging.DEBUG)
    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

from earthkit.data import SimpleFieldList
from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file

LOG = logging.getLogger(__name__)


class WrappedField:
    def __init__(self, field):
        self._field = field

    def __getattr__(self, name):
        return getattr(self._field, name)

    def __repr__(self) -> str:
        return repr(self._field)


class NewDataField(WrappedField):
    def __init__(self, field, data):
        super().__init__(field)
        self._data = data
        self.shape = data.shape

    def to_numpy(self, flatten=False, dtype=None, index=None):
        data = self._data
        if dtype is not None:
            data = data.astype(dtype)
        if flatten:
            data = data.flatten()
        if index is not None:
            data = data[index]
        return data


class NewMetadataField(WrappedField):
    def __init__(self, field, **kwargs):
        super().__init__(field)
        self._metadata = kwargs

    def metadata(self, *args, **kwargs):
        if len(args) == 1 and args[0] in self._metadata:
            return self._metadata[args[0]]
        return self._field.metadata(*args, **kwargs)


def test_ai_models_wrapped_field_grib():

    ds = from_source("file", earthkit_examples_file("test.grib"))
    np_ref = ds.to_numpy()
    vals_ref = ds.values

    f1 = NewMetadataField(ds[0], shortName="sstk")
    f2 = NewMetadataField(ds[1], shortName="2d")

    r = SimpleFieldList([f1, f2])

    assert r[0].metadata("shortName") == "sstk"
    assert r[1].metadata("shortName") == "2d"
    assert np.allclose(r.to_numpy(), np_ref)
    assert np.allclose(r.values, vals_ref)



# (C) Copyright 2020- ECMWF.  #
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import abstractmethod
from collections import defaultdict

LOG = logging.getLogger(__name__)

PRIVATE_ATTRIBUTES = {"observer": lambda: None}


class MetaBase(type):
    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        args, kwargs = cls.patch(obj, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        return obj

    def patch(cls, obj, *args, **kwargs):
        private_attributes = {}
        private_attributes.update(PRIVATE_ATTRIBUTES)
        private_attributes.update(kwargs.pop("_PRIVATE_ATTRIBUTES", {}))

        for k, v in private_attributes.items():
            setattr(obj, k, kwargs.pop(k, v))

        return args, kwargs


class Base(metaclass=MetaBase):
    # Convertors
    def to_numpy(self, **kwargs):
        """Convert into a numpy array"""
        self._not_implemented()

    @abstractmethod
    def to_xarray(self, **kwargs):
        """Convert into an xarray dataset"""
        self._not_implemented()

    @abstractmethod
    def to_pandas(self, **kwargs):
        """Convert into a pandas dataframe"""
        self._not_implemented()

    # Change class
    def mutate(self):
        return self

    @classmethod
    def merge(cls, *args, **kwargs):
        """Merge the object with other ones."""
        return None

    @abstractmethod
    def metadata(self, *args, **kwargs):
        """Return metadata."""
        self._not_implemented()

    # I/O
    @abstractmethod
    def to_target(self, *args, **kwargs):
        """Write data into the specified target."""
        self._not_implemented()

    @abstractmethod
    def datetime(self):
        """Return datetime."""
        self._not_implemented()

    @abstractmethod
    def bounding_box(self):
        """Return the bounding box."""
        self._not_implemented()

    @abstractmethod
    def sel(self, *args, **kwargs):
        """Filter the object based on metadata."""
        self._not_implemented()

    @abstractmethod
    def isel(self, *args, **kwargs):
        self._not_implemented()

    @abstractmethod
    def order_by(self, *args, **kwargs):
        """Reorder the elements of the object."""
        self._not_implemented()

    def unique_values(self, *coords, remapping=None, patches=None, progress_bar=False):
        """Given a list of metadata attributes, such as date, param, levels,
        returns the list of unique values for each attributes
        """
        from earthkit.data.core.order import build_remapping

        assert len(coords)
        assert all(isinstance(k, str) for k in coords), coords

        remapping = build_remapping(remapping, patches)
        iterable = self

        if progress_bar:
            from earthkit.data.utils.progbar import progress_bar

            if progress_bar:
                iterable = progress_bar(
                    iterable=self,
                    desc=f"Finding coords in dataset for {coords}",
                )

        vals = defaultdict(dict)
        for f in iterable:
            metadata = remapping(f.metadata)
            for k in coords:
                v = metadata(k, default=None)
                vals[k][v] = True

        vals = {k: tuple(values.keys()) for k, values in vals.items()}

        return vals

    # @abstractmethod
    # def to_points(self, *args, **kwargs):
    #     self._not_implemented()

    # @abstractmethod
    # def to_latlon(self, *args, **kwargs):
    #     self._not_implemented()

    def __add__(self, other):
        self._not_implemented()

    #
    def _not_implemented(self):
        import inspect

        func = inspect.stack()[1][3]
        module = self.__class__.__module__
        name = self.__class__.__name__

        extra = ""
        if hasattr(self, "path"):
            extra = f" on {self.path}"
        raise NotImplementedError(f"{module}.{name}.{func}(){extra}")

    def batched(self, *args):
        """Return iterator for batches of data"""
        self._not_implemented()

    def group_by(self, *args):
        """Return iterator for batches of data grouped by metadata keys"""
        self._not_implemented()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np
import pytest

from earthkit.data import ArrayField
from earthkit.data.utils.metadata.dict import UserMetadata


def test_array_field_usermetadata_nogeom():
    vals = np.linspace(0, 1, 10)
    meta = UserMetadata(
        {
            "shortName": "test",
            "longName": "Test",
            "date": 20180801,
            "time": 300,
        }
    )

    f = ArrayField(vals, meta)

    assert f.metadata("shortName") == "test"
    assert f.metadata("longName") == "Test"
    assert f.metadata("date") == 20180801
    assert f.metadata("time") == 300
    assert f.datetime()["base_time"] == datetime.datetime(2018, 8, 1, 3, 0)
    assert f.shape == (10,)
    assert np.allclose(f.values, vals)

    with pytest.raises(ValueError):
        f.to_latlon()


@pytest.mark.parametrize("_kwargs", [{}, {"shape": None}, {"shape": (10,)}])
def test_array_field_usermetadata_geom(_kwargs):
    vals = np.linspace(0, 1, 10)
    meta = UserMetadata(
        {
            "shortName": "test",
            "longName": "Test",
            "date": 20180801,
            "time": 300,
            "latitudes": np.linspace(-10.0, 10.0, 10),
            "longitudes": np.linspace(20.0, 40.0, 10),
        },
        **_kwargs,
    )

    f = ArrayField(vals, meta)

    assert f.metadata("shortName") == "test"
    assert f.metadata("longName") == "Test"
    assert f.metadata("date") == 20180801
    assert f.metadata("time") == 300
    assert f.datetime()["base_time"] == datetime.datetime(2018, 8, 1, 3, 0)
    assert f.shape == (10,)
    lat = f.to_latlon()["lat"]
    lon = f.to_latlon()["lon"]
    assert lat.shape == (10,)
    assert lon.shape == (10,)
    assert np.allclose(lat, meta["latitudes"])
    assert np.allclose(lon, meta["longitudes"])
    assert np.allclose(f.values, vals)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)


class PandasMixIn:
    def to_pandas(self, latitude=None, longitude=None, **kwargs):
        import numpy as np
        import pandas as pd

        def ident(x):
            return x

        filter = ident

        def select_point(d):
            idx = np.where((d["lat"] == latitude) & (d["lon"] == longitude))
            return dict(lat=d["lat"][idx], lon=d["lon"][idx], value=d["value"][idx])

        if latitude is not None or longitude is not None:
            filter = select_point

        columns = ("lat", "lon", "value")
        frames = []
        for s in self:
            df = pd.DataFrame.from_dict(filter(dict(zip(columns, s.data(columns, flatten=True)))))
            df["datetime"] = s.datetime()["valid_time"]
            for k, v in s.metadata(namespace="mars").items():
                df[k] = v
            frames.append(df)
        return pd.concat(frames)



import earthkit.data

# Change the location of the user defined cache:
earthkit.data.config.set("user-cache-directory", "/big-disk/earthkit-data-cache")

# Change download timeout
earthkit.data.config.set("url-download-timeout", "1m")

# Multiple values can be set together. The argument list
# can be a dictionary:
earthkit.data.config.set({"url-download-timeout": "1m", "check-out-of-date-urls": True})

# Alternatively, we can use keyword arguments. However, because
# the “-” character is not allowed in variable names in Python we have
# to replace “-” with “_” in all the keyword arguments:
earthkit.data.config.set(url_download_timeout="1m", check_out_of_date_urls=True)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from datetime import timedelta
from functools import cached_property

import numpy as np

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import MetadataAccessor
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.indexing.fieldlist import ClonedFieldCore
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.dates import to_datetime

from .coords import LevelSlice
from .coords import TimeSlice

LOG = logging.getLogger(__name__)


class XArrayFieldGeography(Geography):
    def __init__(self, metadata, ds, variable):
        self.metadata = metadata
        self.variable = variable
        self.ds = ds
        self.north, self.west, self.south, self.east = self.ds.bbox(variable)

    def latitudes(self, dtype=None):
        return self.ds._get_latlon(self.data_array, flatten=True, dtype=dtype)[0]

    def longitudes(self, dtype=None):
        return self.ds._get_latlon(self.data_array, flatten=True, dtype=dtype)[1]

    def x(self, dtype=None):
        return self.ds._get_xy(self.data_array, flatten=True, dtype=dtype)[0]

    def y(self, dtype=None):
        return self.ds._get_xy(self.data_array, flatten=True, dtype=dtype)[1]

    def shape(self):
        _, coords = self.ds._get_xy_coords(self.data_array)
        return tuple([self.data_array.coords[v].size for v in coords])

    def _unique_grid_id(self):
        return self.shape()

    def projection(self):
        from earthkit.data.utils.projections import Projection

        return Projection.from_cf_grid_mapping(**self._grid_mapping().attrs)

    def bounding_box(self):
        return BoundingBox(north=self.north, south=self.south, east=self.east, west=self.west)

    def _grid_mapping(self):
        da = self.data_array
        if "grid_mapping" in da.attrs:
            grid_mapping = self.ds[da.attrs["grid_mapping"]]
        else:
            raise AttributeError("no CF-compliant 'grid_mapping' detected in netCDF attributes")
        return grid_mapping

    def gridspec(self):
        raise NotImplementedError("gridspec is not implemented for netcdf/xarray")

    @property
    def data_array(self):
        return self.ds[self.variable]

    def resolution(self):
        # TODO: implement resolution
        return None

    def mars_grid(self):
        # TODO: implement mars_grid
        return None

    def mars_area(self):
        # TODO: code me
        return [self.north, self.west, self.south, self.east]


class XArrayMetadata(RawMetadata):
    LS_KEYS = ["variable", "level", "valid_datetime", "units"]
    NAMESPACES = [
        "default",
        "mars",
    ]
    MARS_KEYS = ["param", "step", "levelist", "levtype", "number", "date", "time"]

    ACCESSORS = {
        "valid_datetime": ["valid_datetime", "valid_time"],
        "base_datetime": ["base_datetime", "forecast_reference_time", "base_time"],
    }

    def __init__(self, field):
        if not isinstance(field, XArrayField):
            raise TypeError(f"XArrayMetadata: expected field type XArrayField, got {type(field)}")
        self._field = field

        data_array = field._ds[field.variable]
        d = dict(data_array.attrs)

        time = field.non_dim_coords.get("valid_time", field.non_dim_coords.get("time"))
        level = None
        level_type = "sfc"

        for s in field.slices:
            if isinstance(s, TimeSlice):
                time = s.value

            if isinstance(s, LevelSlice):
                level = s.value
                level_type = {"pressure": "pl"}.get(s.name, s.name)

        step = 0
        if time is not None:
            self.time = to_datetime(time)
            if "forecast_reference_time" in field._ds.data_vars:
                forecast_reference_time = field._ds["forecast_reference_time"].data
                assert forecast_reference_time.ndim == 0, forecast_reference_time
                forecast_reference_time = forecast_reference_time.astype("datetime64[s]")
                forecast_reference_time = forecast_reference_time.astype(object)
                step = (time - forecast_reference_time).total_seconds()
                assert step % 3600 == 0, step
                step = int(step // 3600)
                d["step"] = step

            date = self.time - timedelta(hours=step)
            d["date"] = int(date.strftime("%Y%m%d"))
            d["time"] = int(date.strftime("%H%M"))

        else:
            self.time = None

        d["variable"] = self._field.variable
        d["level"] = level
        d["levtype"] = level_type

        super().__init__(d)

    def override(self, *args, **kwargs):
        if not args and not kwargs:
            return self
        return None

    @cached_property
    def geography(self):
        return XArrayFieldGeography(self, self._field._ds, self._field.variable)

    def namespaces(self):
        return self.NAMESPACES

    def as_namespace(self, namespace=None):
        if not isinstance(namespace, str) and namespace is not None:
            raise TypeError("namespace must be a str or None")

        if namespace == "default" or namespace == "" or namespace is None:
            return dict(self)
        elif namespace == "mars":
            return self._as_mars()

    def _as_mars(self):
        return dict(
            param=self["variable"],
            step=self.get("step", None),
            levelist=self["level"],
            levtype=self["levtype"],
            number=None,
            date=self.get("date", None),
            time=self.get("time", None),
        )

    def datetime(self):
        return {
            "base_time": self.base_datetime(),
            "valid_time": self.valid_datetime(),
        }

    def base_datetime(self):
        v = self.valid_datetime()
        if v is not None:
            return v - timedelta(hours=self.get("hour", 0))

    def valid_datetime(self):
        if self.time is not None:
            return to_datetime(self.time)

    @MetadataAccessor(ACCESSORS)
    def get(self, key, default=None, *, raise_on_missing=False, **kwargs):
        if key.startswith("mars."):
            key = key[5:]
            if key not in self.MARS_KEYS:
                if raise_on_missing:
                    raise KeyError(f"Invalid key '{key}' in namespace='mars'")
                else:
                    return default
                # if kwargs.get("raise_on_missing", False):
                #     raise KeyError(f"Invalid key '{key}' in namespace='mars'")
                # else:
                #     return kwargs.get("default", None)

        def _key_name(key):
            if key == "param":
                key = "variable"
            elif key == "levelist":
                key = "level"
            return key

        return super().get(_key_name(key), default=default, raise_on_missing=raise_on_missing, **kwargs)

    def ls_keys(self):
        return self.LS_KEYS

    def data_format(self):
        return "netcdf"


class XArrayField(Field):
    def __init__(self, ds, variable, slices, non_dim_coords):
        super().__init__()
        self._ds = ds
        self._da = ds[variable]

        # self.north, self.west, self.south, self.east = ds.bbox(variable)

        self.variable = variable
        self.slices = slices
        self.non_dim_coords = non_dim_coords
        self.name = self.variable

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.variable},"
            + ",".join([f"{s.name}={s.value}" for s in self.slices])
            + ")"
        )

    @cached_property
    def _metadata(self):
        return XArrayMetadata(self)

    def to_xarray(self):
        da = self._ds[self.variable]
        dims = da.dims
        v = {}
        for s in self.slices:
            if s.is_dimension:
                if s.name in dims:
                    v[s.name] = s.index
        return da.isel(**v)

    def to_pandas(self):
        return self.to_xarray().to_pandas()

    # def _to_numpy(self):
    #     return self.to_xarray().to_numpy()

    def _to_numpy(self):
        dimensions = dict((s.name, s.index) for s in self.slices)
        # values = self.owner.xr_dataset[self.variable].isel(dimensions).values
        values = self._ds[self.variable].isel(dimensions).values
        return values

    def _values(self, dtype=None):
        if dtype is None:
            return self._to_numpy()
        else:
            return self._to_numpy().astype(dtype, copy=False)

    @cached_property
    def grid_mapping(self):
        def tidy(x):
            if isinstance(x, np.ndarray):
                return x.tolist()
            if isinstance(x, (tuple, list)):
                return [tidy(y) for y in x]
            if isinstance(x, dict):
                return {k: tidy(v) for k, v in x.items()}
            return x

        # return tidy(
        #     self.owner.xr_dataset[
        #         self.owner.xr_dataset[self.variable].grid_mapping
        #     ].attrs
        # )

        return tidy(self._ds[self._ds[self.variable].grid_mapping].attrs)

    def clone(self, **kwargs):
        return ClonedXarrayField(self, **kwargs)

    def to_target(self, *args, **kwargs):
        raise NotImplementedError

    def _encode(self, *args, **kwargs):
        raise NotImplementedError


class ClonedXarrayField(ClonedFieldCore, XArrayField):
    def __init__(self, field, **kwargs):
        ClonedFieldCore.__init__(self, field, **kwargs)
        XArrayField.__init__(self, field.ds, field.variable, field.slices, field.non_dim_coords)


class NetCDFMetadata(XArrayMetadata):
    pass


class NetCDFField(XArrayField):
    @cached_property
    def _metadata(self):
        return NetCDFMetadata(self)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1 - equivalent solutions
t = ekd.create_target("file-pattern", "res_{shortName}.grib")
for field in ds:
    t.write(field)

t.close()

ds.to_target("file-pattern", "res_{shortName}.grib")



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data.sources.file import FileSourcePathAndParts
from earthkit.data.utils.parts import SimplePart
from earthkit.data.utils.url import UrlSourcePathAndParts


@pytest.mark.parametrize(
    "paths,parts,expected_paths,expected_parts",
    [
        ("a.grib", None, "a.grib", None),
        (["a.grib"], [None], "a.grib", None),
        ("a.grib", [None], "a.grib", None),
        (["a.grib", None], None, "a.grib", None),
        (["a.grib", [(0, 150)]], None, "a.grib", (SimplePart(0, 150),)),
        (["a.grib", (0, 150)], None, "a.grib", (SimplePart(0, 150),)),
        ("a.grib", [(0, 150)], "a.grib", (SimplePart(0, 150),)),
        ("a.grib", [(0, 150), (150, 100)], "a.grib", (SimplePart(0, 250),)),
        (
            "a.grib",
            [(0, 150), (160, 100)],
            "a.grib",
            (SimplePart(0, 150), SimplePart(160, 100)),
        ),
        (
            "a.grib",
            [(150, 100), (0, 150)],
            ("a.grib", "a.grib"),
            ((SimplePart(150, 100),), (SimplePart(0, 150),)),
        ),
        (["a.grib", "b.grib"], None, ["a.grib", "b.grib"], [None, None]),
        (
            ["a.grib", "b.grib"],
            (0, 150),
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150),), (SimplePart(0, 150),)),
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], "b.grib"],
            None,
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150), SimplePart(200, 20)), None),
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", (0, 100)]],
            None,
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150), SimplePart(200, 20)), ((SimplePart(0, 100),))),
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", [(0, 100)]]],
            None,
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150), SimplePart(200, 20)), ((SimplePart(0, 100),))),
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", [(0, 100), (120, 50)]]],
            None,
            ("a.grib", "b.grib"),
            (
                (SimplePart(0, 150), SimplePart(200, 20)),
                ((SimplePart(0, 100), SimplePart(120, 50))),
            ),
        ),
        (
            ["a.grib", "b.grib"],
            (0, 150),
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150),), (SimplePart(0, 150),)),
        ),
    ],
)
def test_file_parts_prepare(paths, parts, expected_paths, expected_parts):
    p = FileSourcePathAndParts.from_paths(paths, parts)
    res_paths, res_parts = p.path, p.parts
    assert res_paths == expected_paths
    assert res_parts == expected_parts


def test_file_parts_update_1():
    p = FileSourcePathAndParts.from_paths("", None)
    assert p.path == ""
    assert p.parts is None

    p.update("a.grib")
    assert p.path == "a.grib"
    assert p.parts is None

    p.update(["a.grib", "b.grib"])
    assert p.path == ["a.grib", "b.grib"]
    assert p.parts == [None, None]


def test_file_parts_update_2():
    p = FileSourcePathAndParts.from_paths("", None)
    assert p.path == ""
    assert p.parts is None

    p.update(["a.grib"])
    assert p.path == ["a.grib"]
    assert p.parts == [None]

    p.update("a.grib")
    assert p.path == "a.grib"
    assert p.parts is None


def test_file_parts_iter_1():
    p = FileSourcePathAndParts.from_paths("a.grib", None)
    assert p.path == "a.grib"
    assert p.parts is None

    for pt, pr in p:
        pt == "a.grib"
        pr is None


def test_file_parts_iter_2():
    p = FileSourcePathAndParts.from_paths(["a.grib", "b.grib"], None)
    assert p.path == ["a.grib", "b.grib"]
    assert p.parts == [None, None]

    ref_path = ["a.grib", "b.grib"]
    ref_parts = [None, None]

    i = 0
    for pt, pr in p:
        pt == ref_path[i]
        pr is ref_parts[i]
        i += 1


def test_file_parts_zipped_1():
    p = FileSourcePathAndParts.from_paths("a.grib", None)
    assert p.path == "a.grib"
    assert p.parts is None
    assert p.zipped() == [("a.grib", None)]


def test_file_parts_zipped_2():
    p = FileSourcePathAndParts.from_paths(["a.grib", "b.grib"], None)
    assert p.path == ["a.grib", "b.grib"]
    assert p.parts == [None, None]
    assert p.zipped() == [("a.grib", None), ("b.grib", None)]


@pytest.mark.parametrize(
    "urls,parts,expected_values",
    [
        ("a.grib", None, [("a.grib", None)]),
        (["a.grib", None], None, [("a.grib", None)]),
        (["a.grib", [(0, 150)]], None, [("a.grib", (SimplePart(0, 150),))]),
        (["a.grib", (0, 150)], None, [("a.grib", (SimplePart(0, 150),))]),
        ("a.grib", [(0, 150)], [("a.grib", (SimplePart(0, 150),))]),
        ("a.grib", [(0, 150), (150, 100)], [("a.grib", (SimplePart(0, 250),))]),
        (
            "a.grib",
            [(0, 150), (160, 100)],
            [("a.grib", (SimplePart(0, 150), SimplePart(160, 100)))],
        ),
        (
            "a.grib",
            [(150, 100), (0, 150)],
            [
                ("a.grib", ((SimplePart(150, 100),))),
                ("a.grib", ((SimplePart(0, 150),))),
            ],
        ),
        (["a.grib", "b.grib"], None, [("a.grib", None), ("b.grib", None)]),
        (
            ["a.grib", "b.grib"],
            (0, 150),
            [("a.grib", (SimplePart(0, 150),)), ("b.grib", (SimplePart(0, 150),))],
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], "b.grib"],
            None,
            [("a.grib", (SimplePart(0, 150), SimplePart(200, 20))), ("b.grib", None)],
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", (0, 100)]],
            None,
            [
                ("a.grib", (SimplePart(0, 150), SimplePart(200, 20))),
                ("b.grib", (SimplePart(0, 100),)),
            ],
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", [(0, 100)]]],
            None,
            [
                ("a.grib", (SimplePart(0, 150), SimplePart(200, 20))),
                ("b.grib", (SimplePart(0, 100),)),
            ],
        ),
        (
            [["a.grib", [(0, 150), (200, 20)]], ["b.grib", [(0, 100), (120, 50)]]],
            None,
            [
                ("a.grib", (SimplePart(0, 150), SimplePart(200, 20))),
                ("b.grib", (SimplePart(0, 100), SimplePart(120, 50))),
            ],
        ),
        (
            ["a.grib", "b.grib"],
            (0, 150),
            [
                ("a.grib", (SimplePart(0, 150),)),
                ("b.grib", (SimplePart(0, 150),)),
            ],
        ),
    ],
)
def test_url_parts_prepare(urls, parts, expected_values):
    p = UrlSourcePathAndParts.from_paths(urls, parts)
    res = p.zipped()
    assert res == expected_values


@pytest.mark.parametrize(
    "paths,parts,expected_paths,expected_parts,same",
    [
        (["a.grib", "b.grib"], None, ["a.grib", "b.grib"], [None, None], True),
        (["b.grib", "a.grib"], None, ["a.grib", "b.grib"], [None, None], False),
        (
            ["b.grib", "a.grib"],
            (0, 150),
            ("a.grib", "b.grib"),
            ((SimplePart(0, 150),), (SimplePart(0, 150),)),
            False,
        ),
        (
            [["b.grib", [(0, 150), (200, 20)]], "a.grib"],
            None,
            ("a.grib", "b.grib"),
            (None, (SimplePart(0, 150), SimplePart(200, 20))),
            False,
        ),
    ],
)
def test_file_parts_sorted(paths, parts, expected_paths, expected_parts, same):
    p = FileSourcePathAndParts.from_paths(paths, parts)
    p1 = p.sorted()

    if same:
        assert list(p1.path) == list(p.path)
    else:
        assert list(p1.path) != list(p.path)

    res_paths, res_parts = p1.path, p1.parts
    assert list(res_paths) == list(expected_paths)
    assert list(res_parts) == list(expected_parts)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from functools import cached_property

import numpy as np

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.utils.bbox import BoundingBox

from . import Reader


class GeoTIFFGeography(Geography):

    def __init__(self, ds):
        self._ds = ds
        self.x_dim = ds.rio.x_dim
        self.y_dim = ds.rio.y_dim

    def _latlon_coords(self):
        try:
            from pyproj import Transformer
        except ImportError:
            raise ImportError("geotiff handling requires 'pyproj' to be installed")

        return Transformer.from_crs(self._ds.rio.crs, "EPSG:4326", always_xy=True).transform(
            *self._xy_coords()
        )

    def latitudes(self, dtype=None):
        return self._latlon_coords()[0]

    def longitudes(self, dtype=None):
        return self._latlon_coords()[1]

    def _xy_coords(self):
        x = self._ds.coords[self.x_dim].values
        y = self._ds.coords[self.y_dim].values
        return np.meshgrid(x, y)

    def x(self, dtype=None):
        return self._xy_coords()[0]

    def y(self, dtype=None):
        return self._xy_coords()[1]

    def shape(self):
        return self._ds.rio.height, self._ds.rio.width

    def _unique_grid_id(self):
        return (*self.shape(), self._ds.rio.crs.to_wkt())

    def projection(self):
        from earthkit.data.utils.projections import Projection

        return Projection.from_cf_grid_mapping(**self._grid_mapping)

    @property
    def _grid_mapping(self):
        if self._ds.rio.grid_mapping == "spatial_ref":
            return self._ds.coords["spatial_ref"].attrs
        raise NotImplementedError

    def bounding_box(self):
        left, bottom, right, top = self._ds.rio.transform_bounds("EPSG:4326")
        return BoundingBox(north=top, west=left, south=bottom, east=right)

    def gridspec(self):
        raise NotImplementedError

    def resolution(self):
        # Get width and height of pixels in units of the CRS
        x, y = self._ds.rio.resolution()
        crs = self._ds.rio.crs
        units, _ = crs.units_factor
        # Geographic coordinate systems use latitude and longitude
        if crs.is_geographic and units == "degree":
            x = abs(round(x * 1_000_000) / 1_000_000)
            y = abs(round(y * 1_000_000) / 1_000_000)
            if x == y:
                return x
        # raise NotImplementedError(f"resolution for {crs} ({units}, {factor})")

    def mars_grid(self):
        raise NotImplementedError

    def mars_area(self):
        raise NotImplementedError


class GeoTIFFMetadata(RawMetadata):

    LS_KEYS = ["variable", "band"]

    def __init__(self, field, band, geography=None):
        self._field = field
        self._geo = geography
        super().__init__({"variable": field._da.name, "band": band, **field._da.attrs})

    @property
    def geography(self):
        if self._geo is None:
            self._geo = GeoTIFFGeography(self._field._da)
        return self._geo

    def ls_keys(self):
        return self.LS_KEYS


class GeoTIFFField(Field):
    """A GeoTIFF band"""

    def __init__(self, da, band, geography=None):
        super().__init__()
        self._da = da
        self._geo = geography
        self.band = band

    def __repr__(self):
        return f"{self.__class__.__name__}({self.band},{self._da.name})"

    @cached_property
    def _metadata(self):
        return GeoTIFFMetadata(self, self.band, geography=self._geo)

    def to_xarray(self):
        return self._da

    def to_pandas(self):
        # Series with x and y in MultiIndex, not y-Index and x in columns
        series = self._da.to_pandas().stack()
        series.name = self._da.name
        return series

    def _values(self, dtype=None):
        return self._da.values.astype(dtype)

    def write(self, f):
        self._not_implemented()


class GeoTIFFFieldList(FieldList):
    """A list of GeoTIFF bands"""

    FIELD_TYPE = GeoTIFFField

    DEFAULT_XARRAY_KWARGS = {
        # Splitting bands into individual variables preserves band-specific metadata.
        "band_as_variable": True,
        # Mask and scale values by default to match xarray's default behaviour.
        # Note: masked values are set to NaN, so all values are returned as floats.
        "mask_and_scale": True,
        "decode_times": True,
    }

    def __init__(self, path, **kwargs):
        self._fields = None
        self.path = path
        self._ds = self.rioxarray_read()
        # Shared geography instance for all fields/bands
        self._geo = GeoTIFFGeography(self._ds)
        super().__init__(**kwargs)

    def rioxarray_read(self, **kwargs):
        try:
            import rioxarray
        except ImportError:
            raise ImportError("geotiff handling requires 'rioxarray' to be installed")

        options = dict(**self.DEFAULT_XARRAY_KWARGS)
        # Read options from dedicated kwarg if exists, otherwise use all kwargs
        options.update(kwargs.get("rioxarray_open_rasterio_kwargs", kwargs))

        return rioxarray.open_rasterio(self.path, **options)

    def to_xarray(self, **kwargs):
        return self.rioxarray_read(**kwargs)

    def to_pandas(self):
        import pandas as pd

        series = [field.to_pandas() for field in self]
        return pd.concat(series, keys=[s.name for s in series])

    def __len__(self):
        return len(self._ds.data_vars)

    def _getitem(self, n):
        if isinstance(n, int):
            return self.fields[n]

    @property
    def fields(self):
        if self._fields is None:
            self._fields = self._get_fields(self._ds)
        return self._fields

    def _get_fields(self, ds, names=None):
        fields = []
        # Follow GDAL convention and count GeoTIFF bands from 1
        for band, (name, da) in enumerate(ds.data_vars.items(), start=1):
            # Prefer long_name over band_* naming of band_as_variable=True
            if "long_name" in da.attrs:
                da = da.rename(da.attrs["long_name"])
            fields.append(self.FIELD_TYPE(da, band, geography=self._geo))
        return fields

    def describe(self, *args, **kwargs):
        self._not_implemented()


class GeoTIFFReader(GeoTIFFFieldList, Reader):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        GeoTIFFFieldList.__init__(self, path)

    def __repr__(self):
        return f"GeoTIFFReader({self.path})"

    def mutate_source(self):
        return self

    def default_encoder(self):
        return Reader.default_encoder(self)


def _match_magic(magic):
    # https://docs.ogc.org/is/19-008r4/19-008r4.html#_tiff_core_test
    # Bytes 0-1: 'II' (little endian) or 'MM' (big endian)
    # Bytes 2-3: 42 as short in the corresponding byte order
    #           or 43 for a bigtiff file
    # Bytes 4-7: offset to first image file directory
    return magic is not None and len(magic) >= 8 and magic[:4] in {b"II*\x00", b"II+\x00", b"MM\x00*"}


def reader(source, path, *, magic=None, **kwargs):
    if _match_magic(magic):
        return GeoTIFFReader(source, path)



# Test RequestIterStreamer object from sources/url.py

#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.utils.stream import RequestIterStreamer


def iter_stream(chunk_size, data):
    num = len(data)
    pos = 0
    while pos < num:
        start = pos
        end = min(pos + chunk_size, num)
        pos += end - start
        yield data[start:end]


@pytest.mark.parametrize("chunk_size", [1, 2, 3, 4, 5, 10, 12, 14])
@pytest.mark.parametrize("read_size", [1, 2, 3, 4, 5, 10, 12, 14])
def test_request_iter_streamer(chunk_size, read_size):
    data = str.encode("0123456789abc")

    stream = RequestIterStreamer(iter_stream(chunk_size, data))

    assert not stream.closed
    assert stream.read(-2) == bytes()
    assert stream.read(0) == bytes()
    assert stream.peek(4) == data[:4]
    assert not stream.closed

    for i in range(0, len(data), read_size):
        assert stream.read(read_size) == data[i : min(i + read_size, len(data))], i

    assert stream.read(1) == bytes()
    assert stream.closed
    assert stream.read(0) == bytes()
    assert stream.read() == bytes()
    assert stream.read(-1) == bytes()
    assert stream.peek(4) == bytes()


@pytest.mark.parametrize("chunk_size", [1, 2, 3, 4, 5, 10, 12, 14])
def test_request_iter_streamer_read_all_1(chunk_size):
    data = str.encode("0123456789abc")

    stream = RequestIterStreamer(iter_stream(chunk_size, data))

    assert not stream.closed
    assert stream.read(-2) == bytes()
    assert stream.read(0) == bytes()
    assert stream.peek(4) == data[:4]
    assert not stream.closed

    assert stream.read() == data

    assert stream.read(1) == bytes()
    assert stream.closed
    assert stream.read(0) == bytes()
    assert stream.read(-1) == bytes()
    assert stream.peek(4) == bytes()


@pytest.mark.parametrize("chunk_size", [1, 2, 3, 4, 5, 10, 12, 14])
def test_request_iter_streamer_read_all_2(chunk_size):
    data = str.encode("0123456789abc")

    stream = RequestIterStreamer(iter_stream(chunk_size, data))

    assert not stream.closed
    assert stream.read() == data

    assert stream.read(1) == bytes()
    assert stream.closed
    assert stream.read(0) == bytes()
    assert stream.read(-1) == bytes()
    assert stream.peek(4) == bytes()


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import EncodedData
from . import Encoder

LOG = logging.getLogger(__name__)


class CSVEncodedData(EncodedData):
    def __init__(self, df):
        self.df = df

    def to_bytes(self):
        raise NotImplementedError

    def to_file(self, f, **kwargs):
        self.df.to_csv(f, **kwargs)

    def metadata(self, key):
        raise NotImplementedError


class CSVEncoder(Encoder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def encode(
        self,
        data=None,
        **kwargs,
    ):
        if data is not None:
            from earthkit.data.wrappers import get_wrapper

            data = get_wrapper(data)
            return data._encode(self, **kwargs)
        else:
            raise ValueError("No data to encode")

    def _encode(self, data, **kwargs):
        assert data is not None
        return CSVEncodedData(data.to_pandas())

    def _encode_field(self, field, **kwargs):
        return self._encode(field, **kwargs)

    def _encode_fieldlist(self, fieldlist, **kwargs):
        raise NotImplementedError

    def _encode_xarray(self, data, **kwargs):
        raise NotImplementedError


encoder = CSVEncoder



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import datetime

import numpy as np
import pytest

from earthkit.data.utils.metadata.dict import UserMetadata


def test_usermetadata_nogeom():
    meta = UserMetadata(
        {
            "shortName": "test",
            "longName": "Test",
            "date": 20180801,
            "time": 300,
        }
    )

    assert meta["shortName"] == "test"
    assert meta["longName"] == "Test"
    assert meta["date"] == 20180801
    assert meta["time"] == 300
    assert meta["base_datetime"] == "2018-08-01T03:00:00"
    assert meta.base_datetime() == datetime.datetime(2018, 8, 1, 3, 0)
    assert meta.geography.latitudes() is None
    assert meta.geography.longitudes() is None


@pytest.mark.parametrize("_kwargs", [{}, {"shape": None}, {"shape": (10,)}])
def test_usermetadata_geom(_kwargs):
    meta = UserMetadata(
        {
            "shortName": "test",
            "longName": "Test",
            "date": 20180801,
            "time": 300,
            "latitudes": np.linspace(-10.0, 10.0, 10),
            "longitudes": np.linspace(20.0, 40.0, 10),
        },
        **_kwargs,
    )

    assert meta["shortName"] == "test"
    assert meta["longName"] == "Test"
    assert meta["date"] == 20180801
    assert meta["time"] == 300
    assert meta["base_datetime"] == "2018-08-01T03:00:00"
    assert meta.base_datetime() == datetime.datetime(2018, 8, 1, 3, 0)
    assert meta.geography.shape() == (10,)

    lat = meta.geography.latitudes()
    lon = meta.geography.longitudes()
    assert lat.shape == (10,)
    assert lon.shape == (10,)
    assert np.allclose(lat, meta["latitudes"])
    assert np.allclose(lon, meta["longitudes"])


@pytest.mark.parametrize(
    "initial, update, expected",
    [
        ({"shortName": "2t"}, {}, {"shortName": "2t"}),  # No update
        ({"shortName": "2t"}, {"shortName": "msl"}, {"shortName": "msl"}),  # Update existing key
        (
            {"shortName": "2t"},
            {"longName": "Temperature"},
            {"shortName": "2t", "longName": "Temperature"},
        ),  # Add new key
        ({}, {"shortName": "2t"}, {"shortName": "2t"}),  # Add key to empty metadata
        (
            {"shortName": "2t", "longName": "Temperature"},
            {"shortName": "temperature"},
            {"shortName": "temperature", "longName": "Temperature"},
        ),  # Update one key, keep others
    ],
)
def test_usermetadata_override(initial, update, expected):

    initial_copied = copy.deepcopy(initial)
    update_copied = copy.deepcopy(update)

    meta = UserMetadata(initial)
    new_meta = meta.override(update)
    _ = meta.override(**update)  # Check that the override method works with keyword arguments

    # Check that the updated metadata matches the expected result
    for k, v in expected.items():
        assert new_meta[k] == v

    # Ensure the original metadata remains unchanged
    for k, v in initial_copied.items():
        assert meta[k] == v

    # Check that the updated metadata contains all expected keys
    for k in update_copied:
        if k in initial:
            assert meta[k] == initial[k]
        else:
            assert k not in meta


def test_usermetadata_override_shape():
    meta = UserMetadata({}, shape=(10, 1))
    new_meta = meta.override(
        {
            "shortName": "2t",
            "longName": "Temperature",
        }
    )
    new_meta._shape = None
    assert new_meta._shape is None
    assert meta._shape is not None


@pytest.mark.parametrize("data,ref_base", [({"base_datetime": "2018-08-01T09:00:00"}, "2018-08-01T09:00:00")])
def test_usermetadata_base_date_only(data, ref_base):
    meta = UserMetadata(data)

    ref_base_dt = datetime.datetime.fromisoformat(ref_base)
    assert meta["base_datetime"] == ref_base
    assert meta.base_datetime() == ref_base_dt
    assert meta["valid_datetime"] == ref_base
    assert meta.valid_datetime() == ref_base_dt
    assert meta.datetime() == {"base_time": ref_base_dt, "valid_time": ref_base_dt}


@pytest.mark.parametrize(
    "data,ref_base,ref_valid", [({"valid_datetime": "2018-08-01T09:00:00"}, None, "2018-08-01T09:00:00")]
)
def test_usermetadata_valid_date_only(data, ref_base, ref_valid):
    meta = UserMetadata(data)

    ref_valid_dt = datetime.datetime.fromisoformat(ref_valid)
    assert meta["valid_datetime"] == ref_valid
    assert meta.valid_datetime() == ref_valid_dt
    assert meta.datetime() == {"base_time": ref_base, "valid_time": ref_valid_dt}


@pytest.mark.parametrize(
    "data,ref_base,ref_valid,ref_step",
    [
        (
            {
                "base_datetime": "2018-08-01T03:00:00",
                "step": 6,
            },
            "2018-08-01T03:00:00",
            "2018-08-01T09:00:00",
            6,
        ),
        (
            {
                "valid_datetime": "2018-08-01T09:00:00",
                "step": 6,
            },
            "2018-08-01T03:00:00",
            "2018-08-01T09:00:00",
            6,
        ),
        (
            {
                "forecast_reference_time": "2018-08-01T03:00:00",
                "step": 6,
            },
            "2018-08-01T03:00:00",
            "2018-08-01T09:00:00",
            6,
        ),
        (
            {
                "forecast_reference_time": "2018-08-01T03:00:00",
                "step": datetime.timedelta(hours=6),
            },
            "2018-08-01T03:00:00",
            "2018-08-01T09:00:00",
            datetime.timedelta(hours=6),
        ),
        (
            {
                "date": 20180801,
                "time": 300,
                "step": datetime.timedelta(hours=6),
            },
            "2018-08-01T03:00:00",
            "2018-08-01T09:00:00",
            datetime.timedelta(hours=6),
        ),
    ],
)
def test_usermetadata_forecast(data, ref_base, ref_valid, ref_step):
    meta = UserMetadata(data)

    ref_base_dt = datetime.datetime.fromisoformat(ref_base)
    ref_valid_dt = datetime.datetime.fromisoformat(ref_valid)

    if isinstance(ref_step, int):
        ref_step_td = datetime.timedelta(hours=ref_step)
    else:
        ref_step_td = ref_step

    assert meta["valid_datetime"] == ref_valid
    assert meta.valid_datetime() == ref_valid_dt
    assert meta["base_datetime"] == ref_base
    assert meta.base_datetime() == ref_base_dt
    assert meta.datetime() == {"base_time": ref_base_dt, "valid_time": ref_valid_dt}
    assert meta["step"] == ref_step
    assert meta["step_timedelta"] == ref_step_td
    assert meta.step_timedelta() == ref_step_td
    assert meta["forecast_reference_time"] == ref_base


def test_usermetadata_hdate_from_mars():
    meta = UserMetadata(
        {
            "class": "od",
            "expver": "0001",
            "stream": "enfh",
            "type": "cf",
            "levtype": "sfc",
            "param": "167.128",
            "date": "20180830",  # Model version
            "hdate": "20100830",  # Start date of the forecasts
            "time": "0000",  # Forecast starts at 0am
            "step": 12,  # Forecast 12 hours ahead
        }
    )

    ref_base_dt = datetime.datetime(2010, 8, 30, 0, 0)
    ref_valid_dt = datetime.datetime(2010, 8, 30, 12, 0)

    assert meta.base_datetime() == ref_base_dt
    assert meta.valid_datetime() == ref_valid_dt
    assert meta.datetime() == {"base_time": ref_base_dt, "valid_time": ref_valid_dt}
    assert meta["step"] == 12
    assert meta["step_timedelta"] == datetime.timedelta(hours=12)
    assert meta["date"] == "20180830"
    assert meta["hdate"] == "20100830"
    assert meta["time"] == "0000"
    assert meta["forecast_reference_time"] == "2010-08-30T00:00:00"



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
from abc import abstractmethod

import eccodes
from pdbufr.high_level_bufr.bufr import bufr_code_is_coord

from earthkit.data.core import Base
from earthkit.data.core.index import Index
from earthkit.data.core.index import MaskIndex
from earthkit.data.core.index import MultiIndex
from earthkit.data.utils.message import CodesHandle
from earthkit.data.utils.message import CodesMessagePositionIndex
from earthkit.data.utils.message import CodesReader
from earthkit.data.utils.metadata.args import metadata_argument
from earthkit.data.utils.parts import Part
from earthkit.data.utils.summary import make_bufr_html_tree

from .. import Reader
from .pandas import PandasMixIn

BUFR_LS_KEYS = {
    "edition": "edition",
    "type": "dataCategory",
    "subtype": "dataSubCategory",
    "c": "bufrHeaderCentre",
    "mv": "masterTablesVersionNumber",
    "lv": "localTablesVersionNumber",
    "subsets": "numberOfSubsets",
    "compr": "compressedData",
    "typicalDate": "typicalDate",
    "typicalTime": "typicalTime",
    "ident": "ident",
    "lat": "localLatitude",
    "lon": "localLongitude",
}


class BufrCodesMessagePositionIndex(CodesMessagePositionIndex):
    MAGIC = b"BUFR"

    # This does not belong here, should be in the C library
    def _get_message_positions_part(self, fd, part):
        assert part is not None
        assert len(part) == 2

        offset = part[0]
        end_pos = part[0] + part[1] if part[1] > 0 else -1

        if os.lseek(fd, offset, os.SEEK_SET) != offset:
            return

        while True:
            code = os.read(fd, 4)
            if len(code) < 4:
                break

            if code != self.MAGIC:
                offset = os.lseek(fd, offset + 1, os.SEEK_SET)
                continue

            length = self._get_bytes(fd, 3)
            edition = self._get_bytes(fd, 1)

            if end_pos > 0 and offset + length > end_pos:
                return

            if edition in [3, 4]:
                yield offset, length

            offset = os.lseek(fd, offset + length, os.SEEK_SET)


class BUFRCodesHandle(CodesHandle):
    PRODUCT_ID = eccodes.CODES_PRODUCT_BUFR

    def __init__(self, handle, path, offset):
        super().__init__(handle, path, offset)
        self._unpacked = False

    def unpack(self):
        """Decode data section"""
        if not self._unpacked:
            eccodes.codes_set(self._handle, "unpack", 1)
            self._unpacked = True

    def pack(self):
        """Encode data section"""
        if self._unpacked:
            eccodes.codes_set(self._handle, "pack", 1)
            self._unpacked = False

    def json_dump(self, path):
        self.unpack()
        with open(path, "w") as f:
            eccodes.codes_dump(self._handle, f, "json")
        self.pack()

    def __iter__(self):
        class _KeyIterator:
            def __init__(self, handle):
                self._iterator = eccodes.codes_bufr_keys_iterator_new(handle)

            def __del__(self):
                try:
                    eccodes.codes_bufr_keys_iterator_delete(self._iterator)
                except Exception:
                    pass

            def __iter__(self):
                return self

            def __next__(self):
                while True:
                    if not eccodes.codes_bufr_keys_iterator_next(self._iterator):
                        raise StopIteration

                    return eccodes.codes_bufr_keys_iterator_get_name(self._iterator)

        return _KeyIterator(self._handle)

    def keys(self, namespace=None):
        """Iterate over all the available keys"""
        return self.__iter__()

    def as_namespace(self, namespace=None):
        return {k: self.get(k, default=None) for k in self.keys(namespace=namespace)}


class BUFRCodesReader(CodesReader):
    PRODUCT_ID = eccodes.CODES_PRODUCT_BUFR
    HANDLE_TYPE = BUFRCodesHandle


class BUFRMessage(Base):
    r"""Represents a BUFR message in a BUFR file.

    Parameters
    ----------
    path: str
        Path to the BUFR file
    offset: number
        File offset of the message (in bytes)
    length: number
        Size of the message (in bytes)
    """

    def __init__(self, path, offset, length):
        self.path = path
        self._offset = offset
        self._length = length
        self._handle = None

    @property
    def handle(self):
        r""":class:`CodesHandle`: Gets an object providing access to the low level BUFR message structure."""
        if self._handle is None:
            assert self._offset is not None
            self._handle = BUFRCodesReader.from_cache(self.path).at_offset(self._offset)
        return self._handle

    def __repr__(self):
        return "BUFRMessage(type=%s,subType=%s,subsets=%s,%s,%s)" % (
            self.handle.get("dataCategory", default=None),
            self.handle.get("dataSubCategory", default=None),
            self.handle.get("numberOfSubsets", default=None),
            self.handle.get("typicalDate", default=None),
            self.handle.get("typicalTime", default=None),
        )

    def _header(self, key):
        return self.handle.get(key, default=None)

    def subset_count(self):
        """Returns the number of subsets in the given BUFR message."""
        return self._header("numberOfSubsets")

    def is_compressed(self):
        """Checks if the BUFR message contains compressed subsets."""
        return self.subset_count() > 1 and self._header("compressedData") == 1

    def is_uncompressed(self):
        """Checks if the BUFR message contains uncompressed subsets."""
        return self.subset_count() > 1 and self._header("compressedData") == 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._handle = None

    def __setitem__(self, key, value):
        """Sets value associated with ``key``"""
        if isinstance(value, list):
            return eccodes.codes_set_array(self.handle._handle, key, value)
        else:
            return eccodes.codes_set(self.handle._handle, key, value)

    def __getitem__(self, key):
        """Returns the value of the ``key``."""
        return self.handle.get(key)

    def __iter__(self):
        """Returns an iterator for the keys the message contains."""
        return self.handle.__iter__()

    def unpack(self):
        """Decodes the data section of the message. When a message is unpacked all
        the keys in the data section become available via :obj:`metadata`.

        See Also
        --------
        :obj:`unpack`
        """
        return self.handle.unpack()

    def pack(self):
        """Encodes the data section of the message. Having called ``pack`` the
        contents of only the header keys become available via :obj:`metadata`. To access
        the data section you need to use :obj:`unpack` again.

        See Also
        --------
        :obj:`unpack`
        """
        return self.handle.unpack()

    def metadata(self, *keys, astype=None, **kwargs):
        r"""Returns metadata values from the BUFR message. When the message in packed
        (default state) only the header keys are available. To access the data section keys
        you need to call :obj:`unpack`.

        Parameters
        ----------
        *keys: tuple
            Positional arguments specifying metadata keys. Only ecCodes BUFR keys can be used
            here. It can contain a single str or a list or tuple. Can be empty, in this case
            all the keys will
            be used.
        astype: type name, :obj:`list` or :obj:`tuple`
            Return types for ``keys``. A single value is accepted and applied to all the ``keys``.
            Otherwise, must have same the number of elements as ``keys``. Only used when
            ``keys`` is not empty.
        **kwargs: tuple, optional
            Other keyword arguments:

            * default: value, optional
                Specifies the same default value for all the ``keys`` specified. When ``default`` is
                **not present** and a key is not found or its value is a missing value
                :obj:`metadata` will raise KeyError.

        Returns
        -------
        single value, :obj:`list`, :obj:`tuple` or :obj:`dict`
            - when ``keys`` is not empty:
                - single value when ``keys`` is a str
                - otherwise the same type as that of ``keys`` (:obj:`list` or :obj:`tuple`)
            - when ``keys`` is empty:
                - returns a :obj:`dict` with one item per key

        Raises
        ------
        KeyError
            If no ``default`` is set and a key is not found in the message or it has a missing value.

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/temp_10.bufr")
        >>> ds[0].metadata("edition")
        3
        >>> ds[0].metadata("dataCategory", "dataSubCategory")
        (2, 101)

        """
        key, namespace, astype, key_arg_type = metadata_argument(*keys, namespace=None, astype=astype)

        assert isinstance(key, list)
        assert isinstance(namespace, (list, tuple))

        if key:
            assert isinstance(astype, (list, tuple))
            r = [self.handle.get(k, ktype=kt, **kwargs) for k, kt in zip(key, astype)]

            if key_arg_type is str:
                return r[0]
            elif key_arg_type is tuple:
                return tuple(r)
            else:
                return r
        else:
            return self.handle.as_namespace()

    def is_coord(self, key):
        """Check if the specified key is a BUFR coordinate descriptor

        Parameters
        ----------
        key: str
            Key name (can contain ecCodes rank)

        Returns
        -------
        bool
            True if the specified ``key`` is a BUFR coordinate descriptor
        """
        try:
            return bufr_code_is_coord(self.d[key + "->code"])
        except Exception:
            return False

    def dump(self, subset=1):
        r"""Generates a dump with the message content represented as a tree view in a Jupyter notebook.

        Parameters
        ----------
        subset: int
            Subset to dump. Please note that susbset indexing starts at 1. Use None to dump all the
            subsets in the message.

        Returns
        -------
        HTML
            Dump contents represented as a tree view in a Jupyter notebook.

        Examples
        --------
        :ref:`/examples/bufr_temp.ipynb`

        """
        from earthkit.data.core.temporary import temp_file

        with temp_file() as filename:
            self.handle.json_dump(filename)
            with open(filename, "r") as f:
                import json
                import warnings

                try:
                    d = json.loads(f.read())
                    return make_bufr_html_tree(
                        d,
                        self.__repr__(),
                        subset,
                        self.is_compressed(),
                        self.is_uncompressed(),
                    )
                except Exception as e:
                    warnings.warn("Failed to parse bufr_dump", e)
                    return None

    def write(self, f):
        r"""Writes the message to a file object.

        Parameters
        ----------
        f: file object
            The target file object.
        """
        self.handle.write_to(f)

    def message(self):
        r"""Returns a buffer containing the encoded message.

        Returns
        -------
        bytes
        """
        return self.handle.get_buffer()


class BUFRListMixIn(PandasMixIn):
    def ls(self, *args, **kwargs):
        r"""Generates a list like summary of the BUFR message list using a set of metadata keys.

        Parameters
        ----------
        n: int, None
            The number of :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s to be
            listed. ``None`` means all the messages, ``n > 0`` means messages from the front, while
            ``n < 0`` means messages from the back of the list.
        keys: list of str, dict, None
            Metadata keys. To specify a column title for each key in the output use a dict. If
            ``keys`` is None the following dict will be used to define the titles and the keys::

                {
                    "edition": "edition",
                    "type": "dataCategory",
                    "subtype": "dataSubCategory",
                    "c": "bufrHeaderCentre",
                    "mv": "masterTablesVersionNumber",
                    "lv": "localTablesVersionNumber",
                    "subsets": "numberOfSubsets",
                    "compr": "compressedData",
                    "typicalDate": "typicalDate",
                    "typicalTime": "typicalTime",
                    "ident": "ident",
                    "lat": "localLatitude",
                    "lon": "localLongitude",
                }

        extra_keys: list of str, dict, None
            List of additional keys to ``keys``. To specify a column title for each key in the output
            use a dict.

        Returns
        -------
        Pandas DataFrame
            DataFrame with one row per :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`.

        Examples
        --------
        :ref:`/examples/bufr_temp.ipynb`

        """
        from earthkit.data.utils.summary import ls

        def _proc(keys, n):
            count_start = 0
            if n is None:
                count_end = len(self)
            elif n > 0:
                count_end = n
            else:
                num = len(self)
                count_start = max(0, num + n)
                count_end = num

            for count, msg in enumerate(self):
                if count_start <= count < count_end:
                    yield ({k: msg._header(k1) for k, k1 in keys.items()})
                elif count >= count_end:
                    break

        return ls(_proc, BUFR_LS_KEYS, *args, **kwargs)

    def head(self, n=5, **kwargs):
        r"""Generates a list like summary of the first ``n``
        :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0) to be printed from the front.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        Notes
        -----
        The following calls are equivalent:

            .. code-block:: python

                ds.head()
                ds.head(5)
                ds.head(n=5)
                ds.ls(5)
                ds.ls(n=5)

        """
        if n <= 0:
            raise ValueError("head: n must be > 0")
        return self.ls(n=n, **kwargs)

    def tail(self, n=5, **kwargs):
        r"""Generates a list like summary of the last ``n``
        :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``-n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0)  to be printed from the back.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        Notes
        -----
        The following calls are equivalent:

            .. code-block:: python

                ds.tail()
                ds.tail(5)
                ds.tail(n=5)
                ds.ls(-5)
                ds.ls(n=-5)

        """
        if n <= 0:
            raise ValueError("n must be > 0")
        return self.ls(n=-n, **kwargs)

    def metadata(self, *args, **kwargs):
        r"""Returns the metadata values for each message.

        Parameters
        ----------
        *args: tuple
            Positional arguments defining the metadata keys. Passed to
            :obj:`BUFRMessage.metadata() <data.readers.bufr.bufr.BUFRMessage.metadata>`
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`BUFRMessage.metadata() <data.readers.bufr.bufr.BUFRMessage.metadata>`

        Returns
        -------
        list
            List with one item per :obj:`BUFRMessage <data.readers.bufr.bufr.BUFRMessage.metadata>`

        """
        result = []
        for s in self:
            result.append(s.metadata(*args, **kwargs))
        return result


class BUFRList(BUFRListMixIn, Index):
    r"""Represents a list of
    :obj:`BUFRMessage <data.readers.bufr.bufr.BUFRMessage>`\ s.
    """

    def __init__(self, *args, **kwargs):
        Index.__init__(self, *args, **kwargs)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskBUFRList(*args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, BUFRList) for _ in sources)
        return MultiBUFRList(sources)

    def sel(self, *args, **kwargs):
        """Uses header metadata values to select only certain messages from a BUFRList object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the filter condition as dict.
            (See below for details).
        **kwargs: dict, optional
            Other keyword arguments specifying the filter conditions.
            (See below for details).

        Returns
        -------
        object
            Returns a new object with the filtered elements. It contains a view to the data in the
            original object, so no data is copied.

        Notes
        -----
        Filter conditions are specified by a set of **metadata** keys either by a dictionary (in
        ``*args``) or a set of ``**kwargs``. Both single or multiple keys are allowed to use and each
        can specify the following type of filter values:

        - single value::

            ds.sel(dataCategory="2")

        - list of values::

            ds.sel(dataCategory=[1, 2])

        - **slice** of values (defines a **closed interval**, so treated as inclusive of both the start
        and stop values, unlike normal Python indexing)::

            # filter dataCategory between 1 and 4 inclusively
            ds.sel(dataCategory=slice(1,4))
        """
        kwargs.pop("remapping", None)
        return super().sel(*args, remapping=None, **kwargs)

    def isel(self, *args, **kwargs):
        raise NotImplementedError()

    def order_by(self, *args, **kwargs):
        """Changes the order of the messages in a BUFRList object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the metadata keys to perform the ordering on.
            (See below for details)
        **kwargs: dict, optional
            Other keyword arguments specifying the metadata keys to perform the ordering on.
            (See below for details)

        Returns
        -------
        object
            Returns a new object with reordered messages. It contains a view to the data in the
            original object, so no data is copied.

        """
        kwargs.pop("remapping", None)
        return super().order_by(*args, remapping=None, **kwargs)


class MaskBUFRList(BUFRList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiBUFRList(BUFRList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)


class BUFRInFiles(BUFRList):
    def _getitem(self, n):
        if isinstance(n, int):
            part = self.part(n if n >= 0 else len(self) + n)
            return BUFRMessage(part.path, part.offset, part.length)

    def __len__(self):
        return self.number_of_parts()

    @abstractmethod
    def part(self, n):
        self._not_implemented()

    @abstractmethod
    def number_of_parts(self):
        self._not_implemented()


class BUFRInOneFile(BUFRInFiles):
    def __init__(self, path, parts=None):
        self.path = path
        self._file_parts = parts
        self.__positions = None

    @property
    def _positions(self):
        if self.__positions is None:
            self.__positions = BufrCodesMessagePositionIndex(self.path, parts=self._file_parts)
        return self.__positions

    def part(self, n):
        return Part(self.path, self._positions.offsets[n], self._positions.lengths[n])

    def number_of_parts(self):
        return len(self._positions)


class BUFRReader(BUFRInOneFile, Reader):
    """Represents a BUFR file"""

    appendable = True  # BUFR messages can be added to the same file

    def __init__(self, source, path, parts=None):
        Reader.__init__(self, source, path)
        BUFRInOneFile.__init__(self, path, parts=parts)

    def __repr__(self):
        return "BUFRReader(%s)" % (self.path,)

    @classmethod
    def merge(cls, readers):
        assert all(isinstance(s, BUFRReader) for s in readers), readers
        assert len(readers) > 1

        return MultiBUFRList(readers)

    def mutate_source(self):
        # A BUFRReader is a source itself
        return self

    def default_encoder(self):
        return Reader.default_encoder(self)



#!/usr/bin/env python3

# (C) Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file


def test_bufr_iteration():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))
    assert len(ds) == 10
    assert len([x for x in ds]) == 10
    for f in ds:
        assert f.subset_count() == 1


def test_bufr_message_repr():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))
    assert "BUFRMessage" in ds[0].__repr__()


def test_bufr_metadata():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))
    assert ds[0].subset_count() == 1
    assert ds[0].is_compressed() is False
    assert ds[0].is_uncompressed() is False

    assert ds[0].metadata("dataCategory") == 2

    assert ds.metadata("dataCategory") == [2] * 10


def test_bufr_metadata_uncompressed():
    ds = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/synop_multi_subset_uncompressed.bufr"),
    )
    assert len(ds) == 1
    f = ds[0]
    assert f.subset_count() == 12
    assert f.is_compressed() is False
    assert f.is_uncompressed() is True


def test_bufr_metadata_compressed():
    ds = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/ens_multi_subset_compressed.bufr"),
    )
    assert len(ds) == 1
    f = ds[0]
    assert f.subset_count() == 51
    assert f.is_compressed() is True
    assert f.is_uncompressed() is False


def test_bufr_metadata_tc_compressed():
    ds = from_source("url", earthkit_remote_test_data_file("test-data/ens_tropical_cyclon.bufr3"))
    assert len(ds) == 1
    f = ds[0]
    assert f.subset_count() == 3
    assert f.is_compressed() is True
    assert f.is_uncompressed() is False

    for k in f:
        assert k

    for k, v in f.metadata().items():
        assert k


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


CF_PARAMS_TO_KWARGS = {
    "latitude_of_projection_origin": "central_latitude",
    "longitude_of_projection_origin": "central_longitude",
    "longitude_of_central_meridian": "central_longitude",
    "grid_north_pole_latitude": "pole_latitude",
    "grid_north_pole_longitude": "pole_longitude",
    "north_pole_grid_longitude": "central_rotated_longitude",
    "perspective_point_height": "satellite_height",
    "scale_factor_at_projection_origin": "scale_factor",
    "scale_factor_at_central_meridian": "scale_factor",
    "standard_parallel": "standard_parallels",
    "false_easting": "false_easting",
    "false_northing": "false_northing",
}


CF_PARAMS_TO_GLOBE_KWARGS = {
    "semi_major_axis": "semimajor_axis",
    "semi_minor_axis": "semiminor_axis",
    "inverse_flattening": "inverse_flattening",
}


def to_projection_kwargs(parameters):
    kwargs = {
        CF_PARAMS_TO_KWARGS.get(key): value
        for key, value in parameters.items()
        if key in CF_PARAMS_TO_KWARGS and key not in CF_PARAMS_TO_GLOBE_KWARGS
    }

    if not isinstance(kwargs.get("standard_parallels", []), list):
        kwargs["standard_parallels"] = [kwargs["standard_parallels"]]

    globe_kwargs = {
        CF_PARAMS_TO_GLOBE_KWARGS.get(key, key): value
        for key, value in parameters.items()
        if key in CF_PARAMS_TO_GLOBE_KWARGS
    }

    kwargs["globe"] = globe_kwargs

    return kwargs



#!/usr/bin/env python3

# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import os

import pandas as pd
import pytest
import yaml

from earthkit.data.decorators import availability
from earthkit.data.decorators import normalize
from earthkit.data.utils.availability import Availability


@pytest.fixture
def availability_s2s_as_list():
    s2s_config = """
    ecmwf:
        param: ['2t', 'ci', 'gh', 'lsm', 'msl']
        number: 50
        alldates: {start: '2020-01-02', end: '2020-12-31', freq: 'w-thu'}
    eccc:
        param: ['2t', 'ci', 'gh', 'lsm']
        number: 20
        alldates: {start: '2020-01-02', end: '2020-12-31', freq: 'w-thu'}
    ncep:
        param: ['2t', 'ci', 'gh']
        number: 15
        alldates: {start: '2010-01-02', end: '2010-12-31', freq: 'w-thu'}
    """

    config = yaml.safe_load(s2s_config)

    availability_list = []

    for k, v in config.items():
        dic = dict(origin=k)
        dic["date"] = [x.strftime("%Y-%m-%d") for x in pd.date_range(**v["alldates"])][:3]
        dic["number"] = list(range(1, v["number"] + 1))
        dic["param"] = v["param"][:3]
        availability_list.append(dic)
    return availability_list


def parser_for_availability_s2s(v):
    # import pandas as pd

    if "alldates" in v:
        v["alldates"] = v["alldates"]["start"] + "/" + v["alldates"]["end"]
        # not using: v["alldates"]['freq']
        # but this does not work:
        # v["alldates"] = list(pd.date_range(**v["alldates"]))
        v["date"] = v.pop("alldates")

    if "number" in v:
        s, _, e = v["number"].split("/")
        v["number"] = [x for x in range(int(s), int(e) + 1)]

    for remove in [
        "grid",
        "stream",
        "step",
        "stepintervals",
        "level",
        "levelbis",
    ]:
        v.pop(remove, None)

    return v


C0 = [
    {"level": "500", "param": "Z", "step": "24"},
    {"level": "500", "param": "Z", "step": "36"},
    {"level": "500", "param": "Z", "step": "48"},
    {"level": "500", "param": "T", "step": "24"},
    {"level": "500", "param": "T", "step": "36"},
    {"level": "500", "param": "T", "step": "48"},
    {"level": "850", "param": "T", "step": "36"},
    {"level": "850", "param": "T", "step": "48"},
    {"level": "1000", "param": "Z", "step": "24"},
    {"level": "1000", "param": "Z", "step": "48"},
]

C1 = [
    {"level": "500", "param": "a", "step": "24"},
    {"level": "500", "param": "a", "step": "36"},
    {"level": "500", "param": "a", "step": "48"},
    {"level": "500", "param": "b", "step": "24"},
    {"level": "500", "param": "b", "step": "36"},
    {"level": "500", "param": "b", "step": "48"},
    {"level": "850", "param": "b", "step": "36"},
    {"level": "850", "param": "b", "step": "48"},
    {"level": "1000", "param": "a", "step": "24"},
    {"level": "1000", "param": "a", "step": "48"},
]


def level_param_step_no_default(level, param, step):
    return level, param, step


def func_param(param):
    return param


def test_availability_decorator_from_text():
    g = availability(C0)(level_param_step_no_default)
    g(level="1000", param="Z", step="24")
    with pytest.raises(ValueError):
        g(level="1032100", param="Z", step="24")


def test_availability_decorator_from_json_1():
    @availability("availability.json")
    def g(level, param, step):
        return level_param_step_no_default(level, param, step)

    g("1000", "Z", "24")


def test_availability_decorator_from_json_2():
    g = availability("availability.json")(level_param_step_no_default)
    g("1000", "Z", "24")


def test_availability_constructor_from_text():
    avail = Availability(C0)
    print(avail)


def test_availability_constructor_from_json():
    path = os.path.join(os.path.dirname(__file__), "availability.json")
    avail = Availability(path)
    print(avail)


C1 = [
    {"level": "500", "param": "a", "step": "24"},
    {"level": "500", "param": "a", "step": "36"},
    {"level": "500", "param": "a", "step": "48"},
    {"level": "500", "param": "b", "step": "24"},
    {"level": "500", "param": "b", "step": "36"},
    {"level": "500", "param": "b", "step": "48"},
    {"level": "850", "param": "b", "step": "36"},
    {"level": "850", "param": "b", "step": "48"},
    {"level": "1000", "param": "a", "step": "24"},
    {"level": "1000", "param": "a", "step": "48"},
]

C2 = [
    {"level": 500, "param": "a", "step": 24},
    {"level": 500, "param": "a", "step": 36},
    {"level": 500, "param": "a", "step": 48},
    {"level": 500, "param": "b", "step": 24},
    {"level": 500, "param": "b", "step": 36},
    {"level": 500, "param": "b", "step": 48},
    {"level": 850, "param": "b", "step": 36},
    {"level": 850, "param": "b", "step": 48},
    {"level": 1000, "param": "a", "step": 24},
    {"level": 1000, "param": "a", "step": 48},
]


def test_normalize_availability_on_func_1():
    func1 = availability(C1)(level_param_step_no_default)
    func1 = normalize("param", ["a", "b"])(func1)

    assert func1(level=1000, param="a", step="24") == ("1000", "a", "24")
    with pytest.raises(ValueError, match="invalid .*"):
        func1(level=850, param="a", step="24")
    with pytest.raises(ValueError):
        func1(level="1032100", param="a", step="24")


def test_normalize_availability_on_func_2():
    func2 = availability(C2)(level_param_step_no_default)
    func2 = normalize("param", ["a", "b"])(func2)
    # func2 = normalize("level", type=int)(func2)
    # func2 = normalize("step", type=int)(func2)

    # in C2: {"level": 1000, "param": "a", "step": 24},
    assert func2(level=1000, param="a", step=24) == (1000, "a", 24)
    assert func2(level="1000", param="a", step=24) == (1000, "a", 24)
    with pytest.raises(ValueError, match="invalid .*"):
        func2(level="850", param="a", step="24")
    with pytest.raises(ValueError):
        func2(level="1032100", param="a", step="24")


def test_normalize_availability_on_method():
    norm_decorator = normalize("param", ["a", "b"])
    availability_decorator_1 = availability(C1)

    class A:
        @norm_decorator
        @availability_decorator_1
        def method1(self, level, param, step):
            return level, param, step

    assert A().method1(level="1000", param="a", step="24") == ("1000", "a", "24")
    with pytest.raises(ValueError):
        A().method1(level="1032100", param="a", step="24")


def test_avail_norm_1():
    @normalize("param", ["a", "b"])
    @availability(C1)
    def func1(param):
        return param

    @normalize("param", ["unk1", "unk2"])
    @availability(C1)
    def func2(param):
        return param


def test_avail_norm_2():
    @availability(C1)
    @normalize("param", ["a", "b"])
    def func3(level, param, step):
        return level, param, step

    func3("500", "a", "24")
    with pytest.raises(ValueError, match="invalid .*850.*"):
        func3("850", "a", "24")


def test_normalize_duplicates_availability():
    @availability(C1)
    @normalize("param", ["a", "b"])
    @availability(C1)
    def func6(param):
        return param

    assert func6(param="A") == "a"


def test_normalize_duplicates_normalize_1():
    func7 = func_param
    func7 = normalize("param", ["b"])(func7)
    func7 = normalize("param", ["a", "b"])(func7)
    assert func7(param="B") == "b"
    with pytest.raises(ValueError):
        func7(param="a")

    func7a = availability(C1)(func7)
    assert func7a(param="B") == "b"
    with pytest.raises(ValueError):
        func7a(param="a")


def test_normalize_duplicates_normalize_2():
    # latest normalize overwrite "values" of the previous one.
    # TODO: This should perhaps raise a ValueError duplicate normalizer?
    func8 = func_param
    func8 = normalize("param", ["a", "b"])(func8)
    func8 = normalize("param", ["b"])(func8)
    assert func8(param="B") == "b"
    func8(param="a")


def test_avail_norm_3():
    @normalize("param", ["a", "a"])
    @availability(C1)
    def func5(param):
        return param

    assert func5(param="A") == "a"


def test_availability_3():
    @availability(C1)
    def func7(param, step=24):
        return param

    func7("a", step="36")
    with pytest.raises(ValueError, match=r"Invalid value .*"):
        func7(3, step="36")


def test_order_availability_normalize_int():
    decorators = [
        availability(C2),
        normalize("step", type=int, multiple=True),
        normalize("param", type=str, multiple=True),
        normalize("level", type=int, multiple=False),
    ]
    g = level_param_step_no_default
    for order in itertools.permutations(decorators):
        print(order)
        for decorator in order:
            g = decorator(g)
        print("---", g("1000", "a", "24"))
        assert g("1000", "a", "24") == (1000, ["a"], [24])


def test_order_availability_normalize_int_2():
    decorators = [
        normalize("step", type=int, multiple=True),
        normalize("param", type=str, multiple=True),
        normalize("level", type=int, multiple=False),
    ]
    g = level_param_step_no_default
    for order in itertools.permutations(decorators):
        print(order)
        for decorator in order:
            g = decorator(g)
        print("---", g("1000", "a", "24"))
        assert g("1000", "a", "24") == (1000, ["a"], [24])


def test_order_availability_normalize_no_type_int():
    decorators = [
        normalize("step", multiple=True),
        normalize("param", multiple=True),
        normalize("level", multiple=False),
        availability(C2),
    ]
    g = level_param_step_no_default
    for order in itertools.permutations(decorators):
        print(order)
        for decorator in order:
            g = decorator(g)
        print("---", g("1000", "a", "24"))
        assert g("1000", "a", "24") == (1000, ["a"], [24])
        assert g(1000, "a", 24.0) == (1000, ["a"], [24])


def test_order_availability_normalize_no_type_str():
    decorators = [
        normalize("step", multiple=True),
        normalize("param", multiple=True),
        normalize("level", type=str, multiple=False),
        availability(C1),
    ]
    g = level_param_step_no_default
    for order in itertools.permutations(decorators):
        print(order)
        for decorator in order:
            g = decorator(g)
        print("---", g("1000", "a", "24"))
        assert g(1000, "a", 24) == ("1000", ["a"], ["24"])
        assert g("1000", "a", "24") == ("1000", ["a"], ["24"])


def test_s2s_availability_1(availability_s2s_as_list):
    av = Availability(availability_s2s_as_list)

    print(av.tree())

    with pytest.raises(ValueError):
        av.check(number=75)
    av.check(number=7)

    av.check(number=30, origin="ecmwf")
    with pytest.raises(ValueError):
        av.check(number=30, origin="eccc")

    #   av.check(origin='ncep')
    for i in av.flatten():
        print(i)
    # av.check(origin=["unknown"])
    print(av.tree())

    #    av.check(number=1, date=["2010-01-07","2010-01-14"])
    #    av.check(number=1, date=["2010-01-07","2010-01-14"], origin=["unknown"])
    #    av.check(number=1, date=["2010-01-02","2010-01-09"], origin=["ecmwf", "eccc"])

    print("...")
    print(
        av.missing(
            number=1,
            origin=["ecmwf", "ncep"],
            date="2020-01-09",
            # param=['2t', 'ci'],
        ).tree()
    )
    print(
        av.count(
            number=1,
            origin=["ecmwf", "ncep"],
            date="2020-01-09",
        )
    )


@pytest.fixture
def availability_s2s_as_txt():
    return """
    - origin: ecmwf
      fctype: hindcast
      param: ['2t', 'ci', 'gh', 'lsm', 'msl', 'q', 'rsn', 'sm100', 'sm20', 'sp', 'sst', 'st100', 'st20', 't', 'tcc', 'tcw', 'tp', 'ttr', 'u', 'v']
      number: 1/to/50
      stream: enfh
      step: 0/to/1104/by/24
      stepintervals: 0-24/24-48/48-72/72-96/96-120/120-144/144-168/168-192/192-216/216-240/240-264/264-288/288-312/312-336/336-360/360-384/384-408/408-432/432-456/456-480/480-504/504-528/528-552/552-576/576-600/600-624/624-648/648-672/672-696/696-720/720-744/744-768/768-792/792-816/816-840/840-864/864-888/888-912/912-936/936-960/960-984/984-1008/1008-1032/1032-1056/1056-1080/1080-1104
      levels: 1000/925/850/700/500/300/200/100/50/10
      levelsbis: 1000/925/850/700/300/500/200
      grid: null
      hdate: ALL
      alldates: {start: '2020-01-02', end: '2020-12-31', freq: 'w-thu'}

    - origin: eccc
      fctype: hindcast
      param: ['2t', 'ci', 'gh', 'lsm', 'msl', 'q', 'rsn', 'sp', 'sst', 't', 'tcc', 'tcw', 'tp', 'ttr', 'u', 'v']
      number: 1/to/50
      stream: enfh
      step: 24/to/768/by/24
      stepintervals: 0-24/24-48/48-72/72-96/96-120/120-144/144-168/168-192/192-216/216-240/240-264/264-288/288-312/312-336/336-360/360-384/384-408/408-432/432-456/456-480/480-504/504-528/528-552/552-576/576-600/600-624/624-648/648-672/672-696/696-720/720-744/744-768
      levels: 1000/925/850/700/500/300/200/100/50/10
      levelsbis: 1000/925/850/700/300/500/200
      grid: null
      hdate: ALL
      alldates: {start: '2020-01-02', end: '2020-12-31', freq: 'w-thu'}

    # ncep hindcast has run only once, with date = 2011-03-01
    - origin: ncep
      fctype: hindcast
      param: ['2t', 'ci', 'gh', 'lsm', 'msl', 'q', 'sm100', 'sm20', 'sp', 'sst', 'st100', 'st20', 't', 'tcc', 'tcw', 'tp', 'ttr', 'u', 'v']
      number: 1/to/3
      stream: enfh
      step: 24/to/1056/by/24
      stepintervals: 24-48/48-72/72-96/96-120/120-144/144-168/168-192/192-216/216-240/240-264/264-288/288-312/312-336/336-360/360-384/384-408/408-432/432-456/456-480/480-504/504-528/528-552/552-576/576-600/600-624/624-648/648-672/672-696/696-720/720-744/744-768/768-792/792-816/816-840/840-864/864-888/888-912/912-936/936-960/960-984/984-1008/1008-1032/1032-1056
      levels: 1000/925/850/700/500/300/200/100/50/10
      levelsbis: 1000/925/850/700/300/500/200
      grid: null
      # note that this is 2010, that is why the date for ncep is not starting on 2010-01-02 (which is not a thursday btw)
      alldates: {start: '2010-01-07', end: '2010-12-29', freq: 'w-thu'}
    """  # noqa: E501


def test_s2s_availability_2(availability_s2s_as_txt):
    av = Availability(
        availability_s2s_as_txt,
        parser=parser_for_availability_s2s,
        intervals="date",
    )
    av.check(number=30, origin="eccc")
    with pytest.raises(ValueError):
        av.check(number=100, fctype="hindcast", origin="ecmwf")
    av.check(number=30, date="2020-01-02", origin="eccc")
    with pytest.raises(ValueError):
        av.check(number=30, date="2111-01-02", origin="eccc")

    # this should raise a ValueError
    # with pytest.raises(ValueError):
    #     av.check(number=30, alldates='2020-01-03', origin="eccc")


if __name__ == "__main__":
    # test_order_avaibility_normalize_4()
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class FilePartStreamReader:
    def __init__(self, path, parts, chunk_size=1024 * 1024):
        self.path = path
        self.parts = parts
        self.chunk_size = chunk_size

    def __iter__(self):
        yield from self._iter_chunks()

    # TODO: use this method alternatively
    # def __iter_full_parts(self):
    #     with open(self.path, "rb") as f:
    #         for offset, length in self.parts:
    #             f.seek(offset)
    #             yield f.read(length)

    def _iter_chunks(self):
        with open(self.path, "rb") as f:
            for offset, length in self.parts:
                pos = offset
                size = length
                while size > 0:
                    f.seek(pos)
                    chunk = f.read(min(self.chunk_size, size))
                    if len(chunk) >= size:
                        yield chunk[:size]
                        size = -1
                    else:
                        yield chunk
                        size -= len(chunk)
                        pos += len(chunk)


class RequestIterStreamer:
    """Expose chunk-based stream reader as a stream supporting a generic read method."""

    def __init__(self, iter_content):
        from collections import deque

        self.iter_content = iter_content
        self.content = deque()
        self.position = 0
        self.total = 0
        self.consumed = False

    def _ensure_content(self, size):
        while self.total < size:
            try:
                self.content.append(next(self.iter_content))
                self.total += len(self.content[-1])
            except StopIteration:
                break

    def _read(self, size):
        assert len(self.content) > 0

        start = self.position
        length = min(len(self.content[0]) - start, size)
        end = start + length
        data = self.content[0][start:end]
        last = 0
        size -= length

        if size > 0:
            d = [data]
            for i in range(1, len(self.content)):
                start = 0
                length = min(len(self.content[i]) - start, size)
                end = start + length
                d.append(self.content[i][start:end])
                last = i
                size -= length
                if size <= 0:
                    break
            data = data = b"".join(d)

        return data, last, end, size

    def read(self, size=-1):
        if size < -1 or size == 0 or self.consumed:
            return bytes()

        if size == -1:
            return self.readall()

        self._ensure_content(size)
        if len(self.content) == 0 or self.total == 0:
            self.close()
            return bytes()

        data, last, self.position, missing_size = self._read(size)
        # LOG.debug(f"{size=} {last=} pos={self.position} {missing_size=}")
        if missing_size > 0:
            self.close()
        else:
            if self.position == len(self.content[last]):
                last += 1
                self.position = 0

            if last > 0:
                for _ in range(0, last):
                    self.content.popleft()

            self.total = sum(len(x) for x in self.content)
            self.total -= self.position

        return data

    def readall(self):
        if self.consumed:
            return bytes()

        first = self.read(self.total)
        if len(first) == 0:
            first = next(self.iter_content)
        res = [first]

        for d in self.iter_content:
            res.append(d)

        self.close()

        if len(res) == 1:
            return res[0]
        else:
            return b"".join(res)

    def peek(self, size):
        if size <= 0 or self.consumed:
            return bytes()

        self._ensure_content(size)
        data, _, _, _ = self._read(size)
        return data

    def close(self):
        if not self.closed:
            self._clear()

    @property
    def closed(self):
        return self.consumed

    def _clear(self):
        self.iter_content = None
        self.content.clear()
        self.position = 0
        self.consumed = True



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_EOD


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_EOD, reason="No access to EOD")
def test_eod():
    ds = from_source(
        "ecmwf-open-data",
        param=["2t", "msl"],
        levtype="sfc",
    )
    assert len(ds) == 2
    assert ds.metadata("param") == ["2t", "msl"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_EOD, reason="No access to EOD")
@pytest.mark.parametrize("model", ["ifs", "aifs"])
def test_eod_model(model):
    ds = from_source(
        "ecmwf-open-data",
        model=model,
        param="t",
        levelist=500,
    )
    assert len(ds) == 1
    assert ds[0].metadata(["param", "level"]) == ["t", 500]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file


@pytest.mark.long_test
@pytest.mark.download
def test_netcdf_to_xarray_args():
    # The JD variable in the NetCDF is defined as follows:
    #
    # short JD(time, lat, lon) ;
    #   string JD:long_name = "Date of the first detection" ;
    #   string JD:units = "days since 2022-01-01" ;
    #   string JD:comment = "Possible values: 0  when the pixel is not burned; 1 to 366 day of
    #       the first detection when the pixel is burned; -1 when the pixel is not observed
    #       in the month; -2 when pixel is not burnable: water bodies, bare areas, urban areas,
    #       and permanent snow and ice.
    #
    # when loaded with xarray.open_dataset/xarray.open_mdataset without any kwargs the
    # type of the JD variable is datetime64[ns], which is wrong. The correct type should
    # be int16.

    ds = from_source(
        "url",
        earthkit_remote_test_data_file("test-data", "20220401-C3S-L3S_FIRE-BA-OLCI-AREA_3-fv1.1.nc"),
    )

    r = ds.to_xarray(xarray_open_mfdataset_kwargs=dict(decode_cf=False, decode_times=False))
    assert r["JD"].dtype == "int16"
    r["JD"].shape == (1, 20880, 28440)
    assert np.isclose(r["JD"].values.min(), -2)
    assert np.isclose(r["JD"].values.max(), 120)

    r = ds.to_xarray(decode_cf=False, decode_times=False)
    assert r["JD"].dtype == "int16"
    r["JD"].shape == (1, 20880, 28440)
    assert np.isclose(r["JD"].values.min(), -2)
    assert np.isclose(r["JD"].values.max(), 120)

    r = ds.to_xarray()
    assert r["JD"].dtype == "<M8[ns]"
    r["JD"].shape == (1, 20880, 28440)


if __name__ == "__main__":
    from earthkit.data.testing import main

    # test_datetime()
    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def download_example_file(file_names, remote_dir="examples", force=False):
    import os
    import urllib.request

    from earthkit.data.testing import earthkit_remote_test_data_file

    if isinstance(file_names, str):
        file_names = [file_names]

    for f_name in file_names:
        if force or not os.path.exists(f_name):
            urllib.request.urlretrieve(
                earthkit_remote_test_data_file(os.path.join(remote_dir, f_name)), f_name
            )


def remote_example_file(file_name, remote_dir="examples"):
    import os

    from earthkit.data.testing import earthkit_remote_test_data_file

    return earthkit_remote_test_data_file(os.path.join(remote_dir, file_name))



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils import projections


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


@pytest.mark.parametrize(
    "dtype,expected_dtype",
    [(None, np.float64), (np.float32, np.float32), (np.float64, np.float64)],
)
def test_netcdf_to_points_1(dtype, expected_dtype):
    ds = from_source("file", earthkit_test_data_file("test_single.nc"))

    eps = 1e-5

    v = ds[0].to_points(flatten=True, dtype=dtype)
    assert isinstance(v, dict)
    assert isinstance(v["x"], np.ndarray)
    assert isinstance(v["y"], np.ndarray)
    assert v["x"].dtype == expected_dtype
    assert v["y"].dtype == expected_dtype
    check_array(
        v["x"],
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        v["y"],
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )

    # fieldlist
    v = ds.to_points(flatten=True, dtype=dtype)
    assert isinstance(v, dict)
    assert isinstance(v["x"], np.ndarray)
    assert isinstance(v["y"], np.ndarray)
    assert v["x"].dtype == expected_dtype
    assert v["y"].dtype == expected_dtype


def test_netcdf_to_points_2():
    ds = from_source("file", earthkit_examples_file("test.nc"))

    assert len(ds) == 2

    import xarray as xr

    xr_ds = xr.open_dataset(earthkit_examples_file("test.nc"))

    for f in ds:
        v = f.to_points()
        assert isinstance(v, dict)

        # x
        assert isinstance(v["x"], np.ndarray)
        assert v["x"].shape == (11, 19)
        for x in v["x"]:
            assert np.allclose(x, np.arange(-27, 45 + 4, 4))

        # y
        assert isinstance(v["y"], np.ndarray)
        assert v["y"].shape == (11, 19)
        for i, y in enumerate(v["y"]):
            assert np.allclose(y, np.ones(19) * (73 - i * 4))

        ref = xr_ds[f.name].sel(latitude=57, longitude=-7).values

        x = 5
        y = 4
        assert np.isclose(f.to_numpy()[y, x], ref)
        assert np.isclose(v["x"][y, x], -7)
        assert np.isclose(v["y"][y, x], 57)


def test_netcdf_to_latlon():
    ds = from_source("file", earthkit_examples_file("test.nc"))

    assert len(ds) == 2

    import xarray as xr

    xr_ds = xr.open_dataset(earthkit_examples_file("test.nc"))

    for f in ds:
        v = f.to_latlon()
        assert isinstance(v, dict)

        # lon
        assert isinstance(v["lon"], np.ndarray)
        assert v["lon"].shape == (11, 19)
        for x in v["lon"]:
            assert np.allclose(x, np.arange(-27, 45 + 4, 4))

        # lat
        assert isinstance(v["lat"], np.ndarray)
        assert v["lat"].shape == (11, 19)
        for i, y in enumerate(v["lat"]):
            assert np.allclose(y, np.ones(19) * (73 - i * 4))

        ref = xr_ds[f.name].sel(latitude=57, longitude=-7).values

        x = 5
        y = 4
        assert np.isclose(f.to_numpy()[y, x], ref)
        assert np.isclose(v["lon"][y, x], -7)
        assert np.isclose(v["lat"][y, x], 57)


def test_netcdf_bbox():
    ds = from_source("file", earthkit_examples_file("test.nc"))
    bb = ds.bounding_box()
    assert len(bb) == 2
    for b in bb:
        assert b.as_tuple() == (73, -27, 33, 45)


def test_netcdf_mars_area():
    ds = from_source("file", earthkit_examples_file("test.nc"))
    ref = [73, -27, 33, 45]
    assert np.allclose(np.asarray(ds[0].mars_area), np.asarray(ref))


def test_netcdf_proj_string_non_cf():
    f = from_source("file", earthkit_examples_file("test.nc"))
    with pytest.raises(AttributeError):
        f[0].projection()


def test_netcdf_projection_laea():
    f = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))
    projection = f[0].projection()
    assert isinstance(projection, projections.LambertAzimuthalEqualArea)
    assert projection.parameters == {
        "central_latitude": 52.0,
        "central_longitude": 10.0,
        "false_northing": 3210000.0,
        "false_easting": 4321000.0,
    }
    assert projection.globe == {
        "ellipse": "GRS80",
    }


def test_netcdf_proj_string_laea():
    f = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))
    r = f[0].projection()
    assert (
        r.to_proj_string()
        == "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 +ellps=GRS80 +units=m +no_defs"
    )


def test_netcdf_to_points_laea():
    ds = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))

    assert len(ds) == 3

    pos = [(0, 0), (0, -1), (-1, 0), (-1, -1)]

    # we must check multiple fields
    for idx in range(2):
        v = ds[idx].to_points()
        assert isinstance(v, dict)

        # lon
        assert isinstance(v["x"], np.ndarray)
        assert v["x"].shape == (950, 1000)

        ref = np.array([2502500.0, 7497500.0, 2502500.0, 7497500.0])
        for i, x in enumerate(pos):
            assert np.isclose(v["x"][x], ref[i]), f"{i=}, {x=}"

        # lat
        assert isinstance(v["y"], np.ndarray)
        assert v["y"].shape == (950, 1000)

        ref = np.array([5497500.0, 5497500.0, 752500.0, 752500.0])
        for i, x in enumerate(pos):
            assert np.isclose(v["y"][x], ref[i]), f"{i=}, {x=}"


def test_netcdf_to_latlon_laea():
    ds = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))

    assert len(ds) == 3

    pos = [(0, 0), (0, -1), (-1, 0), (-1, -1)]

    # we must check multiple fields
    for idx in range(2):
        v = ds[idx].to_latlon()
        assert isinstance(v, dict)

        # lon
        assert isinstance(v["lon"], np.ndarray)
        assert v["lon"].shape == (950, 1000)

        ref = np.array(
            [
                -35.034023999999995,
                73.93767587613708,
                -8.229274420493763,
                41.13970495087975,
            ]
        )
        for i, x in enumerate(pos):
            assert np.isclose(v["lon"][x], ref[i]), f"{i=}, {x=}"

        # lat
        assert isinstance(v["lat"], np.ndarray)
        assert v["lat"].shape == (950, 1000)

        ref = np.array(
            [
                66.9821429989222,
                58.24673887576243,
                27.802844211251625,
                23.942342882929605,
            ]
        )
        for i, x in enumerate(pos):
            assert np.isclose(v["lat"][x], ref[i]), f"{i=}, {x=}"


def test_netcdf_forecast_reference_time():
    ds = from_source("url", earthkit_remote_test_data_file("test-data", "fa_ta850.nc"))

    assert len(ds) == 37
    assert ds[0].metadata("valid_datetime") == "2020-01-23T00:00:00"
    assert ds[5].metadata("valid_datetime") == "2020-01-23T05:00:00"


if __name__ == "__main__":
    from earthkit.data.testing import main

    # test_datetime()
    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data.utils.parts import SimplePart
from earthkit.data.utils.url import UrlSpec


@pytest.mark.parametrize(
    "urls,_kwargs,expected_url,expected_parts,expected_items",
    [
        (
            {"url": "a.grib"},
            {},
            ["a.grib"],
            [None],
            [
                (
                    "a.grib",
                    None,
                    {},
                )
            ],
        ),
        (
            {"url": "a.grib", "parts": None},
            {},
            ["a.grib"],
            [None],
            [
                (
                    "a.grib",
                    None,
                    {},
                )
            ],
        ),
        (
            {"url": "a.grib", "parts": (0, 2)},
            {"chunk_size": 12},
            ("a.grib",),
            ((SimplePart(offset=0, length=2),),),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                )
            ],
        ),
        (
            {
                "url": "a.grib",
                "chunk_size": 12,
            },
            {"parts": (0, 2)},
            ("a.grib",),
            ((SimplePart(offset=0, length=2),),),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                )
            ],
        ),
        (
            {"url": "a.grib", "parts": [(0, 2), (5, 3)]},
            {"chunk_size": 12},
            ("a.grib",),
            ((SimplePart(offset=0, length=2), SimplePart(offset=5, length=3)),),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2), SimplePart(offset=5, length=3)),
                    {"chunk_size": 12},
                )
            ],
        ),
        (
            [{"url": "a.grib", "parts": (0, 2)}, {"url": "b.grib"}],
            {"chunk_size": 12},
            ("a.grib", "b.grib"),
            ((SimplePart(offset=0, length=2),), None),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                ),
                (
                    "b.grib",
                    None,
                    {"chunk_size": 12},
                ),
            ],
        ),
        (
            [{"url": "a.grib", "chunk_size": 12}, {"url": "b.grib"}],
            {"parts": (0, 2)},
            ("a.grib", "b.grib"),
            ((SimplePart(offset=0, length=2),), (SimplePart(offset=0, length=2),)),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                ),
                (
                    "b.grib",
                    (SimplePart(offset=0, length=2),),
                    {},
                ),
            ],
        ),
        (
            [
                {"url": "a.grib", "chunk_size": 12},
                {"url": "b.grib"},
                "c.grib",
            ],
            {"parts": (0, 2)},
            ("a.grib", "b.grib", "c.grib"),
            (
                (SimplePart(offset=0, length=2),),
                (SimplePart(offset=0, length=2),),
                (SimplePart(offset=0, length=2),),
            ),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                ),
                (
                    "b.grib",
                    (SimplePart(offset=0, length=2),),
                    {},
                ),
                (
                    "c.grib",
                    (SimplePart(offset=0, length=2),),
                    {},
                ),
            ],
        ),
        (
            [
                {"url": "a.grib", "chunk_size": 12, "parts": (0, 2)},
                {"url": "b.grib"},
                "c.grib",
                ["d.grib", ((0, 5), (9, 3))],
            ],
            {"timeout": 12},
            ("a.grib", "b.grib", "c.grib", "d.grib"),
            (
                (SimplePart(offset=0, length=2),),
                None,
                None,
                (SimplePart(offset=0, length=5), SimplePart(offset=9, length=3)),
            ),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12, "timeout": 12},
                ),
                (
                    "b.grib",
                    None,
                    {"timeout": 12},
                ),
                (
                    "c.grib",
                    None,
                    {"timeout": 12},
                ),
                (
                    "d.grib",
                    (SimplePart(offset=0, length=5), SimplePart(offset=9, length=3)),
                    {"timeout": 12},
                ),
            ],
        ),
    ],
)
def test_url_spec(urls, _kwargs, expected_url, expected_parts, expected_items):
    r = UrlSpec.from_urls(urls, **_kwargs)

    assert r.url == expected_url
    assert r.parts == expected_parts

    for i, x in enumerate(r):
        assert x == expected_items[i]


@pytest.mark.parametrize(
    "urls,_kwargs,expected_url,expected_parts,expected_items",
    [
        (
            {"url": "a.grib", "parts": (0, 2)},
            {"chunk_size": 1, "parts": (3, 5)},
            ("a.grib",),
            ((SimplePart(offset=0, length=2),),),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                )
            ],
        ),
    ],
)
def test_url_spec_bad(urls, _kwargs, expected_url, expected_parts, expected_items):
    with pytest.raises(ValueError):
        UrlSpec.from_urls(urls, **_kwargs)


@pytest.mark.parametrize(
    "urls,_kwargs,expected_url,expected_parts,expected_items",
    [
        (
            [{"url": "b.grib", "chunk_size": 12}, {"url": "a.grib"}],
            {"parts": (0, 2)},
            ("a.grib", "b.grib"),
            ((SimplePart(offset=0, length=2),), (SimplePart(offset=0, length=2),)),
            [
                (
                    "a.grib",
                    (SimplePart(offset=0, length=2),),
                    {},
                ),
                (
                    "b.grib",
                    (SimplePart(offset=0, length=2),),
                    {"chunk_size": 12},
                ),
            ],
        ),
    ],
)
def test_url_spec_sorted(urls, _kwargs, expected_url, expected_parts, expected_items):
    r = UrlSpec.from_urls(urls, **_kwargs)
    r1 = r.sorted()

    assert r1.url != r.url
    # assert r1.parts != r.parts
    assert r1.url == expected_url
    assert r1.parts == expected_parts

    for i, x in enumerate(r1):
        assert x == expected_items[i]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np
import pytest

from earthkit.data.utils.dates import date_to_grib
from earthkit.data.utils.dates import datetime_to_grib
from earthkit.data.utils.dates import mars_like_date_list
from earthkit.data.utils.dates import numpy_datetime_to_datetime
from earthkit.data.utils.dates import numpy_timedelta_to_timedelta
from earthkit.data.utils.dates import step_to_grib
from earthkit.data.utils.dates import time_to_grib
from earthkit.data.utils.dates import timedeltas_to_int
from earthkit.data.utils.dates import to_datetime
from earthkit.data.utils.dates import to_datetime_list
from earthkit.data.utils.dates import to_time
from earthkit.data.utils.dates import to_timedelta

# Change to utc once aware datetime objects will be used
# tzinfo = datetime.timezone.utc
tzinfo = None


def relative_date(n):
    """Since it is based on calling now() returns both the relative date and day after."""
    if n <= 0:
        d = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=n)
        d = datetime.datetime(d.year, d.month, d.day)
        return (d, d + datetime.timedelta(days=1))
    else:
        raise ValueError(f"{n=} must be negative")


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        (20020502, datetime.datetime(2002, 5, 2), None),
        (np.int64(20020502), datetime.datetime(2002, 5, 2), None),
        ("20020502", datetime.datetime(2002, 5, 2), None),
        ("2002-05-02", datetime.datetime(2002, 5, 2), None),
        ("2002-05-02T00", datetime.datetime(2002, 5, 2), None),
        ("2002-05-02T00Z", datetime.datetime(2002, 5, 2, tzinfo=datetime.timezone.utc), None),
        ("2002-05-02T06", datetime.datetime(2002, 5, 2, 6), None),
        ("2002-05-02T06:11", datetime.datetime(2002, 5, 2, 6, 11), None),
        ("2002-05-02T06:11:03", datetime.datetime(2002, 5, 2, 6, 11, 3), None),
        (datetime.datetime(2002, 5, 2, 6, 11, 3), datetime.datetime(2002, 5, 2, 6, 11, 3), None),
        (np.datetime64("2002-05-02"), datetime.datetime(2002, 5, 2, tzinfo=tzinfo), None),
        (np.datetime64(0, "Y"), datetime.datetime(1970, 1, 1, tzinfo=tzinfo), None),
        (0, relative_date(0), None),
        (-1, relative_date(-1), None),
        (1, None, ValueError),
        (20020, None, ValueError),
    ],
)
def test_to_datetime(d, expected_value, error):
    if error is None:
        if isinstance(expected_value, tuple):
            assert to_datetime(d) in expected_value
        else:
            assert to_datetime(d) == expected_value
    else:
        with pytest.raises(error):
            to_datetime(d)


@pytest.mark.parametrize(
    "args,expected_value,error",
    [
        (
            (datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 4), 1),
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3), datetime.datetime(2002, 5, 4)],
            None,
        ),
        (
            (datetime.datetime(2002, 5, 4), datetime.datetime(2002, 5, 2), 1),
            None,
            AssertionError,
        ),
        (
            (datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 4), 0),
            None,
            AssertionError,
        ),
        (
            (datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 4), -1),
            None,
            AssertionError,
        ),
    ],
)
def test_mars_like_date_list(args, expected_value, error):
    if error is None:
        assert mars_like_date_list(*args) == expected_value
    else:
        with pytest.raises(error):
            mars_like_date_list(*args)


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        ([20020502, "to", 20020503], [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)], None),
        ((20020502, "to", 20020503), [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)], None),
        ([20020502, "TO", 20020503], [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)], None),
        (
            ["20020502", "to", "20020503"],
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)],
            None,
        ),
        (
            ["2002-05-02", "to", "2002-05-03"],
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)],
            None,
        ),
        (
            ["2002-05-02T06", "to", "2002-05-03T06"],
            [datetime.datetime(2002, 5, 2, 6), datetime.datetime(2002, 5, 3, 6)],
            None,
        ),
        (
            [datetime.datetime(2002, 5, 2), "to", datetime.datetime(2002, 5, 3)],
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 3)],
            None,
        ),
        (
            [20020502, "to", 20020504, "by", 2],
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 4)],
            None,
        ),
        (
            [datetime.datetime(2002, 5, 2), "to", datetime.datetime(2002, 5, 4), "by", 2],
            [datetime.datetime(2002, 5, 2), datetime.datetime(2002, 5, 4)],
            None,
        ),
        (20020502, [datetime.datetime(2002, 5, 2)], None),
        ("20020502", [datetime.datetime(2002, 5, 2)], None),
        (datetime.datetime(2002, 5, 2), [datetime.datetime(2002, 5, 2)], None),
        (np.datetime64("2002-05-02"), [datetime.datetime(2002, 5, 2, tzinfo=tzinfo)], None),
    ],
)
def test_to_datetime_list(d, expected_value, error):
    if error is None:
        assert to_datetime_list(d) == expected_value
    else:
        with pytest.raises(error):
            to_datetime_list(d)


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        (20020502, datetime.time(0), None),
        (np.int64(20020502), datetime.time(0), None),
        ("20020502", datetime.time(0), None),
        ("2002-05-02", datetime.time(0), None),
        ("2002-05-02T00", datetime.time(0), None),
        ("2002-05-02T00Z", datetime.time(0), None),
        ("2002-05-02T06", datetime.time(6), None),
        ("2002-05-02T06:11", datetime.time(6, 11), None),
        ("2002-05-02T06:11:03", datetime.time(6, 11, 3), None),
        (datetime.datetime(2002, 5, 2, 6, 11, 3), datetime.time(6, 11, 3), None),
        (np.datetime64("2002-05-02"), datetime.time(0), None),
        (np.datetime64(6, "h"), datetime.time(6), None),
    ],
)
def test_to_time(d, expected_value, error):
    if error is None:
        assert to_time(d) == expected_value
    else:
        with pytest.raises(error):
            to_time(d)


@pytest.mark.parametrize(
    "step,expected_delta,error",
    [
        (12, datetime.timedelta(hours=12), None),
        ("12h", datetime.timedelta(hours=12), None),
        ("12s", datetime.timedelta(seconds=12), None),
        ("12m", datetime.timedelta(minutes=12), None),
        ("1m", datetime.timedelta(minutes=1), None),
        ("", None, (ValueError, TypeError)),
        ("m", None, (ValueError, TypeError, AttributeError)),
        ("1Z", None, (ValueError, TypeError)),
        ("m1", None, (ValueError, TypeError)),
        ("-1", None, (ValueError, TypeError)),
        ("-1s", None, (ValueError, TypeError)),
        ("1.1s", None, (ValueError, TypeError)),
    ],
)
def test_to_timedelta(step, expected_delta, error):
    if error is None:
        assert to_timedelta(step) == expected_delta
    else:
        with pytest.raises(error):
            to_timedelta(step)


@pytest.mark.parametrize(
    "td,expected_delta,error",
    [
        (np.timedelta64(61, "s"), datetime.timedelta(minutes=1, seconds=1), None),
        (
            np.timedelta64((2 * 3600 + 61) * 1000, "ms"),
            datetime.timedelta(hours=2, minutes=1, seconds=1),
            None,
        ),
        (
            np.timedelta64((2 * 3600 + 61) * 1000 * 1000 * 1000, "ns"),
            datetime.timedelta(hours=2, minutes=1, seconds=1),
            None,
        ),
    ],
)
def test_numpy_timedelta_to_timedelta(td, expected_delta, error):
    if error is None:
        assert numpy_timedelta_to_timedelta(td) == expected_delta
    else:
        with pytest.raises(error):
            numpy_timedelta_to_timedelta(td)


@pytest.mark.parametrize(
    "td,expected_delta,error",
    [
        (np.datetime64("2002-05-02"), datetime.datetime(2002, 5, 2, tzinfo=tzinfo), None),
        (
            np.datetime64("2002-05-02T06"),
            datetime.datetime(2002, 5, 2, 6, tzinfo=tzinfo),
            None,
        ),
        (
            np.datetime64("2002-05-02T06:23"),
            datetime.datetime(2002, 5, 2, 6, 23, tzinfo=tzinfo),
            None,
        ),
        (np.datetime64(0, "s"), datetime.datetime(1970, 1, 1, tzinfo=tzinfo), None),
        (np.datetime64(30, "s"), datetime.datetime(1970, 1, 1, 0, 0, 30, tzinfo=tzinfo), None),
        (np.datetime64(30, "m"), datetime.datetime(1970, 1, 1, 0, 30, tzinfo=tzinfo), None),
        (np.datetime64(30, "h"), datetime.datetime(1970, 1, 2, 6, tzinfo=tzinfo), None),
    ],
)
def test_numpy_datetime_to_datetime(td, expected_delta, error):
    if error is None:
        assert numpy_datetime_to_datetime(td) == expected_delta
    else:
        with pytest.raises(error):
            numpy_datetime_to_datetime(td)


@pytest.mark.parametrize(
    "td,expected_value,error",
    [
        (
            [datetime.timedelta(hours=12), datetime.timedelta(hours=18)],
            ([12, 18], datetime.timedelta(hours=1)),
            None,
        ),
        (
            [datetime.timedelta(hours=12), datetime.timedelta(hours=18, minutes=30)],
            ([12 * 60, 18 * 60 + 30], datetime.timedelta(minutes=1)),
            None,
        ),
        (
            [datetime.timedelta(hours=12), datetime.timedelta(hours=18, seconds=30)],
            ([12 * 3600, 18 * 3600 + 30], datetime.timedelta(seconds=1)),
            None,
        ),
        (
            datetime.timedelta(hours=12),
            ([12], datetime.timedelta(hours=1)),
            None,
        ),
    ],
)
def test_timedeltas_to_int(td, expected_value, error):
    if error is None:
        assert timedeltas_to_int(td) == expected_value
    else:
        with pytest.raises(error):
            timedeltas_to_int(td)


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        (20020502, 20020502, None),
        (np.int64(20020502), 20020502, None),
        ("20020502", 20020502, None),
        ("2002-05-02", 20020502, None),
        ("2002-05-02T00", 20020502, None),
        ("2002-05-02T00Z", 20020502, None),
        ("2002-05-02T06", 20020502, None),
        ("2002-05-02T06:11", 20020502, None),
        ("2002-05-02T06:11:03", 20020502, None),
        (datetime.datetime(2002, 5, 2, 6, 11, 3), 20020502, None),
        (np.datetime64("2002-05-02"), 20020502, None),
        (np.datetime64(0, "Y"), 19700101, None),
    ],
)
def test_date_to_grib(d, expected_value, error):
    if error is None:
        assert date_to_grib(d) == expected_value
    else:
        with pytest.raises(error):
            date_to_grib(d)


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        (0, 0, None),
        (6, 6, None),
        (12, 12, None),
        (600, 600, None),
        (1200, 1200, None),
        (1230, 1230, None),
        (np.int64(0), 0, None),
        (np.int64(6), 6, None),
        (np.int64(12), 12, None),
        (np.int64(600), 600, None),
        (np.int64(1200), 1200, None),
        (np.int64(1230), 1230, None),
        ("0", 0, None),
        ("6", 6, None),
        ("12", 12, None),
        ("600", 600, None),
        ("1200", 1200, None),
        ("1230", 1230, None),
        (datetime.timedelta(minutes=0), 0, None),
        (datetime.timedelta(minutes=6), 6, None),
        (datetime.timedelta(hours=0), 0, None),
        (datetime.timedelta(hours=6), 600, None),
        (datetime.timedelta(hours=12), 1200, None),
        (datetime.timedelta(hours=12, minutes=6), 1206, None),
        (datetime.timedelta(hours=120, seconds=2), None, ValueError),
        (np.timedelta64(0, "m"), 0, None),
        (np.timedelta64(6, "m"), 6, None),
        (np.timedelta64(0, "h"), 0, None),
        (np.timedelta64(6, "h"), 600, None),
        (np.timedelta64(12, "h"), 1200, None),
        (np.timedelta64(12 * 60 + 30, "m"), 1230, None),
    ],
)
def test_time_to_grib(d, expected_value, error):
    if error is None:
        assert time_to_grib(d) == expected_value
    else:
        with pytest.raises(error):
            time_to_grib(d)


@pytest.mark.parametrize(
    "step,expected_value,error",
    [
        (0, 0, None),
        (6, 6, None),
        (12, 12, None),
        (120, 120, None),
        ("0h", "0h", None),
        ("6h", "6h", None),
        ("12h", "12h", None),
        ("120h", "120h", None),
        ("6m", "6m", None),
        ("6s", "6s", None),
        (np.timedelta64(6, "h"), 6, None),
        (np.timedelta64(6 * 3600 * 1000, "ms"), 6, None),
        (np.timedelta64(6 * 3600 * 1000 * 1000 * 1000, "ns"), 6, None),
        (np.timedelta64(61, "s"), "61s", None),
        (np.timedelta64((2 * 3600 + 61) * 1000, "ms"), "7261s", None),
        (
            np.timedelta64((2 * 3600 + 61) * 1000 * 1000 * 1000, "ns"),
            "7261s",
            None,
        ),
    ],
)
def test_step_to_grib(step, expected_value, error):
    if error is None:
        assert step_to_grib(step) == expected_value
    else:
        with pytest.raises(error):
            step_to_grib(step)


@pytest.mark.parametrize(
    "d,expected_value,error",
    [
        (datetime.datetime(2002, 5, 2), (20020502, 0), None),
        (datetime.datetime(2002, 5, 2, 6), (20020502, 600), None),
        (datetime.datetime(2002, 5, 2, 12), (20020502, 1200), None),
        (np.int64(20020502), (20020502, 0), None),
        ("20020502", (20020502, 0), None),
        ("2002-05-02", (20020502, 0), None),
        ("2002-05-02T00", (20020502, 0), None),
        ("2002-05-02T00Z", (20020502, 0), None),
        ("2002-05-02T06", (20020502, 600), None),
        ("2002-05-02T06:11", (20020502, 611), None),
        ("2002-05-02T06:11:03", (20020502, 611), None),
        (datetime.datetime(2002, 5, 2, 6, 11, 3), (20020502, 611), None),
        (np.datetime64("2002-05-02"), (20020502, 0), None),
        (np.datetime64("2002-05-02T06"), (20020502, 600), None),
        (np.datetime64(0, "Y"), (19700101, 0), None),
    ],
)
def test_datetime_to_grib(d, expected_value, error):
    if error is None:
        assert datetime_to_grib(d) == expected_value
    else:
        with pytest.raises(error):
            datetime_to_grib(d)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from abc import ABCMeta
from abc import abstractmethod


class Iter(metaclass=ABCMeta):
    def __init__(self, data, create=None):
        self.data = data
        self.create = create

    @staticmethod
    def _create(obj, batch):
        if hasattr(obj, "from_list"):
            return obj.from_list
        elif hasattr(obj, "from_fields"):
            return obj.from_fields
        elif len(batch) > 0 and hasattr(batch[0], "to_fieldlist"):
            return batch[0].to_fieldlist

        return lambda x: x

    @abstractmethod
    def _iterator(self, data):
        pass

    @abstractmethod
    def _from_batch(self, obj, batch):
        pass

    @abstractmethod
    def _metadata(self, data, keys):
        pass

    def batched(self, n):
        if n < 1:
            raise ValueError("n must be at least one")

        from itertools import islice

        it = self._iterator(self.data)
        while batch := tuple(islice(it, n)):
            yield self._from_batch(self.data, batch)

    def group_by(self, *args, sort=True):
        keys = self._flatten(args)

        r = self.data.order_by(*keys) if sort else self.data

        from itertools import groupby

        it = self._iterator(r)
        for batch in groupby(it, self._metadata(r, keys)):
            batch = list(batch[1])
            yield self._from_batch(r, batch)

    def _flatten(self, keys):
        _keys = []
        for v in keys:
            if isinstance(v, str):
                _keys.append(v)
            elif isinstance(v, (tuple, list)):
                _keys.extend(v)
        return _keys


class BasicIter(Iter):
    def _iterator(self, data):
        return iter(data)

    def _from_batch(self, obj, batch):
        if self.create is None:
            self.create = self._create(obj, batch)
        return self.create(batch)

    def _metadata(self, data, keys):
        return lambda f: f._attributes(keys)


class IndexedIter(Iter):
    def _iterator(self, data):
        print(f"{data=} {len(data)}")
        return iter(range(len(data)))

    def _from_batch(self, obj, batch):
        if len(batch) >= 2:
            batch = slice(batch[0], batch[-1] + 1)
        return obj[batch]

    def _metadata(self, data, keys):
        return lambda idx: data[idx]._attributes(keys)


def batched(data, n, mode="iter", create=None):
    it = _ITERS.get(mode, None)
    if it is not None:
        return it(data, create=create).batched(n)
    else:
        raise ValueError(f"invalid mode={mode}")


def group_by(data, *args, mode="iter", sort=True, create=None):
    it = _ITERS.get(mode, None)
    if it is not None:
        return it(data, create=create).group_by(*args, sort=sort)
    else:
        raise ValueError(f"invalid mode={mode}")


_ITERS = {"iter": BasicIter, "indexed": IndexedIter}



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from collections import namedtuple

from earthkit.data.core.constants import DATETIME
from earthkit.data.core.order import build_remapping
from earthkit.data.core.order import normalize_order_by
from earthkit.data.core.select import normalize_selection
from earthkit.data.decorators import cached_method
from earthkit.data.decorators import normalize
from earthkit.data.indexing.database.sql import SqlDatabase
from earthkit.data.indexing.database.sql import SqlOrder
from earthkit.data.indexing.database.sql import SqlRemapping
from earthkit.data.indexing.database.sql import SqlSelection
from earthkit.data.readers.grib.index.db import FieldListInFilesWithDBIndex
from earthkit.data.utils.serialise import register_serialisation

LOG = logging.getLogger(__name__)

SqlResultCache = namedtuple("SqlResultCache", ["first", "length", "result"])


@normalize(DATETIME, "date-list", format="%Y-%m-%d %H:%M:%S")
def _normalize_grib_kwargs_values(**kwargs):
    return kwargs


class FieldListInFilesWithSqlIndex(FieldListInFilesWithDBIndex):
    DBCLASS = SqlDatabase
    DB_CACHE_SIZE = 100_000
    DB_DICT_CACHE_SIZE = 100_000

    def apply_filters(self, filters):
        obj = self
        for f in filters:
            obj = obj.filter(f)
        return obj

    def _find_all_indices_dict(self):
        from earthkit.data.indexing.database import GRIB_KEYS_NAMES

        d = self.unique_values(*GRIB_KEYS_NAMES, remapping=None, progress_bar=None)

        for k, v in d.items():
            try:
                v.remove(None)
            except ValueError:
                pass
            d[k] = sorted(v)

        return d

    def unique_values(self, *coords, remapping=None, progress_bar=None):
        """Given a list of metadata attributes, such as date, param, levels,
        returns the list of unique values for each attributes
        """
        keys = coords

        remapping = build_remapping(remapping)
        # print("Not using remapping here")

        coords = {k: None for k in coords}
        coords = list(coords.keys())
        # print("coords:", coords)
        values = self.db.unique_values(*coords, remapping=remapping).values()

        dic = {k: v for k, v in zip(keys, values)}
        return dic

    def filter(self, filter):
        if filter.is_empty:
            return self
        db = self.db.filter(filter)
        return self.__class__(db=db)

    def sel(self, *args, remapping=None, **kwargs):
        kwargs = normalize_selection(*args, **kwargs)
        if DATETIME in kwargs and kwargs[DATETIME] is not None:
            kwargs = _normalize_grib_kwargs_values(**kwargs)

        if not kwargs:
            return self
        return self.filter(SqlSelection(kwargs, remapping))

    def order_by(self, *args, remapping=None, **kwargs):
        kwargs = normalize_order_by(*args, **kwargs)

        out = self

        if remapping is not None:
            out = out.filter(SqlRemapping(remapping=remapping))

        if kwargs:
            out = out.filter(SqlOrder(kwargs))

        return out

    def part(self, n):
        if self._cache is None or not (self._cache.first <= n < self._cache.first + self._cache.length):
            first = (n // self.DB_CACHE_SIZE) * self.DB_CACHE_SIZE
            result = self.db.lookup_parts(limit=self.DB_CACHE_SIZE, offset=first)
            self._cache = SqlResultCache(first, len(result), result)
        return self._cache.result[n % self.DB_CACHE_SIZE]

    def get_metadata(self, n):
        assert "Used only in virtual"
        if self._dict_cache is None or not (
            self._dict_cache.first <= n < self._dict_cache.first + self._dict_cache.length
        ):
            first = (n // self.DB_DICT_CACHE_SIZE) * self.DB_DICT_CACHE_SIZE
            result = self.db.lookup_dicts(
                limit=self.DB_DICT_CACHE_SIZE,
                offset=first,
                with_parts=False,
                # remove_none=False ?
            )
            result = list(result)

            self._dict_cache = SqlResultCache(first, len(result), result)
        return self._dict_cache.result[n % self.DB_DICT_CACHE_SIZE]

    @cached_method
    def number_of_parts(self):
        return self.db.count()


register_serialisation(
    FieldListInFilesWithSqlIndex,
    lambda x: [x.db.db_path, x.db._filters],
    lambda x: FieldListInFilesWithSqlIndex(db=SqlDatabase(x[0])).apply_filters(filters=x[1]),
)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

try:
    import hda
except ImportError:
    raise ImportError("WEkEO access requires 'hda' to be installed")

import yaml

from earthkit.data.core.thread import SoftThreadPool

from .file import FileSource
from .prompt import APIKeyPrompt

LOG = logging.getLogger(__name__)


class HDAAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://www.wekeo.eu"
    retrieve_api_key_url = "https://www.wekeo.eu"

    prompts = [
        dict(
            name="user",
            example="name",
            title="User name",
            hidden=False,
            validate=r"[0-9a-z]+",
        ),
        dict(
            name="password",
            example="secretpassword",
            title="Password",
            hidden=True,
            validate=r"[0-9A-z\!\@\#\$\%\&\*]{5,30}",
        ),
    ]

    rcfile = "~/.hdarc"
    rcfile_env = "HDA_RC"
    config_env = ("HDA_USER", "HDA_PASSWORD")

    def save(self, input, file):
        yaml.dump(input, file, default_flow_style=False)

    def load(self, file):
        return yaml.safe_load(file.read())


class ApiClient(hda.Client):
    name = "wekeo"

    def __int__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

    def retrieve(self, name, request, target=None):
        matches = self.search(request["request"])
        out = []
        for result in matches.results:
            self.accept_tac(matches.dataset)
            download_id = matches._get_download_id(result)
            out.append(os.path.abspath(self.stream(download_id, result["properties"]["size"], target)))
        return out


EXTENSIONS = {
    "grib": ".grib",
    "netcdf": ".nc",
}


class WekeoRetriever(FileSource):
    sphinxdoc = """
    WekeoRetriever
    """

    def __init__(self, dataset, *args, prompt=True, **kwargs):
        super().__init__()

        self.prompt = prompt

        assert isinstance(dataset, str)
        if len(args):
            assert len(args) == 1
            assert isinstance(args[0], dict)
            assert not kwargs
            kwargs = args[0]

        requests = self.requests(**kwargs)

        self.client(self.prompt)  # Trigger password prompt before threading

        nthreads = min(self.config("number-of-download-threads"), len(requests))

        if nthreads < 2:
            self.path = [self._retrieve(dataset, r) for r in requests]
        else:
            from earthkit.data.utils.progbar import tqdm

            with SoftThreadPool(nthreads=nthreads) as pool:
                futures = [pool.submit(self._retrieve, dataset, r) for r in requests]

                iterator = (f.result() for f in futures)
                self.path = list(tqdm(iterator, leave=True, total=len(requests)))

    @staticmethod
    def client(use_prompt):
        if use_prompt:
            prompt = HDAAPIKeyPrompt()
            prompt.check()

            try:
                return ApiClient()
            except Exception as e:
                # if no rc file is available hda throws
                # ConfigurationError: Missing or incomplete configuration
                if ".hdarc" in str(e) or not prompt.has_config_env():
                    LOG.warning(e)
                    LOG.exception(f"Could not load hda client. {e}")
                    prompt.ask_user_and_save()
                    return ApiClient()
                raise
        else:
            return ApiClient()

    def _retrieve(self, dataset, request):
        def retrieve(target, args):
            self.client(self.prompt).retrieve(args[0], args[1], target)

        return self.cache_file(
            retrieve,
            (dataset, request),
            extension=EXTENSIONS.get(request.get("format"), ".cache"),
        )

    @staticmethod
    def requests(**kwargs):
        split_on = kwargs.pop("split_on", None)
        if split_on is None or not isinstance(kwargs.get(split_on), (list, tuple)):
            return [kwargs]

        result = []

        for v in kwargs[split_on]:
            r = dict(**kwargs)
            r[split_on] = v
            result.append(r)

        return result


source = WekeoRetriever
